"""
"""
from logging import getLogger

from naludaq.communication import ControlRegisters, DigitalRegisters

from .default import BoardController

LOGGER = getLogger("naludaq.board_controller_trbhm")


class TrbhmBoardController(BoardController):
    """Special board controller for TRBHM."""

    def start_readout(
        self,
        trig="self",
        lb="trigrel",
        acq="raw",
        dig_head=False,
        ped="zero",
        readoutEn=True,
        singleEv=False,
    ):
        """Sends a readout signal to the board.

        Obstensibly reads out the chip?

        Args:
            trig (str):
                'immediate', '00' - digitize whatever instantly as fast as possible.
                'ext': '01' - use the external trigger.
                'self': '10' - trigger on input RF signal.
            lbType (str):
                'forced': '00' - always reads out same set of windows, start window set by lookback
                'trig': '01' - relative to the trigger.
                'roi': '10' - complicated...
            acq_type (str):
                'raw': '00' - Can't ped-subtract. <Default output>
                'ped': '01' - sub, spits out headers, channels, and windows. <Requires firmware modification>
            dig_head (bool):
                True: reads out a "digitization" header after each sram read (only works with acq_type="ped")
                False: doesn't read out that header
            pedMode (str):
                'z': '00' - all peds = 0
                'c': '01' - peds = chan*1024, set offset per channel
                'r': '11' - peds = RAM addr
            readoutEn (bool): Enables readout to ethernet (otherwise data thrown away)
                self.readReadoutRate() is the companion to this to see the true rate
                readout rate scalar stored in FPGA reg N_GPR(32) + 5
            singleEV (bool): True - only reads one event.

        Command
        -------

        command is: BMMXXVVV
        MM = singleEvent(7) + readoutEn(6) + pedSel(5 dt 4) + modeSel(3 dt 0)
        - modeSel=x"0" analog write
        - modeSel=x"1" digital write
        - modeSel=x"2" digital read
        - modeSel=x"3" waveform read

        VVVVV = addr_sel(19 dt 12) + data_sel(11 dt 0)
        - data_sel(5 downto 4) = trig type
        - data_sel(3 downto 2) = lb type
        - data_sel(1 downto 0) = acq type

        """
        readout_params = self._validate_readout_params(
            trig, lb, acq, dig_head, ped, readoutEn, singleEv
        )

        ControlRegisters(self.board).write("exttrig", True)
        readout_params["trig"] = "ext"
        if trig[0] == "e":  # ext
            trig_sel = 0b00
        elif trig[0] == "s":  # self
            trig_sel = 0b01
        else:  # imm
            trig_sel = 0b10
        ControlRegisters(self.board).write("trig_sel", trig_sel)

        ControlRegisters(self.board).write("fifo_unload_or_event_en", True)

        cmd = self._generate_start_readout_cmd(readout_params)
        LOGGER.debug("READOUT command: %s", cmd)

        self._send_command(cmd)
        self.is_reading_out = True

    def stop_readout(self):
        super().stop_readout()
        ControlRegisters(self.board).write("fifo_unload_or_event_en", False)

    def read_scalar(self, channel: int) -> int:
        """Read the scalar for the given channel"""
        channels_per_chip = self.board.channels // self.board.available_chips
        relative_channel = channel % channels_per_chip
        chip = channel // channels_per_chip
        return self._read_scalar_inner(chip, relative_channel)

    def _read_scalar_inner(self, chip: int, relative_channel: int) -> int:
        """Read the scalar for the given channel.

        This is the inner function that actually reads the scalar.
        It is called by `read_scalar` and should not be called directly.
        """
        name = self.get_scal_name(relative_channel)
        scal = self._read_digital_register(name, chips=chip)
        try:
            scalhigh = self._read_digital_register("scalhigh", chips=chip)
        except (KeyError, AttributeError):
            scalhigh = 0
        shift_amt = DigitalRegisters(self.board).registers[name]["bitwidth"]
        scal += scalhigh << shift_amt

        return scal

    # def _toggle_trigger(self):
    #     """Toggles the ext trigger using software.

    #     For TRBHM the wait between separate register writes is too long, and
    #     toggling the trigger too slowly results in too many events coming back
    #     and filling the FIFO, causing malformed events. This method instead
    #     sends the register writes all as one string.
    #     """
    #     cr = ControlRegisters(self.board)

    #     wait_cmd = "AE000001"
    #     exttrig_high_cmd = cr.generate_write("exttrig", True)
    #     exttrig_low_cmd = cr.generate_write("exttrig", False)
    #     toggle_cmd = wait_cmd + exttrig_high_cmd + exttrig_low_cmd
    #     self._send_command(toggle_cmd)

    def set_loopback_enabled(self, enabled: bool):
        """Set whether serial loopback is enabled.

        Loopback can safely be disabled during most of the operations with the board.
        Loopback **must** be disabled when communicating over the serial interface.
        If serial communication with the ASIC is intended then this should run during startup and only be enabled as needed.

        Args:
            enabled (bool): True to enable loopback.

        Raises:
            TypeError if enabled is not a bool.
        """
        if not isinstance(enabled, bool):
            raise TypeError("Argument must be bool")
        OFF = "B0900002"
        ON = "B0900003"
        cmd = ON if enabled else OFF
        self._send_command(cmd)



"""Package creator is the second part in a multithread double buffer.
The packager slices teh raw data from the serial reader into packages.

OVERVIEW:
Checks the applications serial buffer for data.
Reads all data in the buffer and splits the data into packages.

Return package:
{'created_at': time.time(),
 'length': length of data,
 raw_data': bytesarray()
}

"""
import logging
import time
from collections import defaultdict, deque
from typing import List

from naludaq.daq.workers.packager.worker_packager import PackagerLight

logger = logging.getLogger(__name__)


UNUSED_TIMING = -1_000_000  # This must be sufficently far from timing margin.
TIMING_MARGIN = 1500


class HDSoCPackager(PackagerLight):
    """A specialized version of the `PackagerLight` for HDSoCv1.

    The packager will split the data in the ``input_buffer`` into
    ``output_buffer`` and ``output_answers`` depending on the pkg length.

    The HDSoC has a different timing system than the other boards.
    The timing system is used to sort the packets into events.
    The timing is a 24-bit number, it's stored in the header of the packet.

    To sort the packages correctly into events the timing of each packet is parsed
    Then it's compared to the timing of the previous packet.
    If the timing is within the margin, the packet is added to the event.
    If the timing is outside the margin, the previous event is completed and a new event is started.

    The packager also has a timeout to separate data originating from two
    different readouts. This prevents the problem of mixing old and new data
    and offsetting later data by some number of windows.

    Args:
        board (Board): the board object
        input_buffer (deque): incoming stream from serial reader, raw bytes
        output_buffer (deque): output for raw event `dict`s.
        output_answers (deque): answers received from serial reader during readout
        stop_word (bytes): Stop word to slice on
        frequency (int): How quick to poll the input buffer and slice

    """

    def __init__(
        self,
        board,
        input_buffer: deque,
        output_buffer: deque,
        output_answers: deque,
        stop_word: bytes,
        frequency: int,
    ):
        super().__init__(
            input_buffer=input_buffer,
            output_buffer=output_buffer,
            output_answers=output_answers,
            stop_word=stop_word,
            frequency=frequency,
        )
        self._board = board
        self._chan_mask = board.params.get("chanmask", 0x3F)
        self._package_size = board.params.get("package_size", 72)

        self._timing_mask = board.params.get("timing_mask", 0xFFF)
        self._timing_shift = board.params.get("timing_shift", 12)

        self._create_incomplete_event(UNUSED_TIMING)

        self.timing_margin = TIMING_MARGIN
        self._do_timeout = True
        self._timeout = 0.8  # determined experimentally
        self._last_packet_time = 0

    def _slice_buffer(self, internal_buffer: bytearray):
        """Slice the internal buffer at the stopword, and
        separates the data into events, answers, and remaining (incomplete)
        data.

        This specialized HDSoC version of the function builds events from
        received packets over multiple calls, and returns the built
        events once they are complete.

        Args:
            internal_buffer (bytearray): raw data to slice.

        Returns:
            A tuple of (output, answers, rest):
                - output: a list of rawevent packages(dict)
                - answers: a list of answer packages
                - rest: whatever is after the last stopword.
        """
        complete_events = []
        raw_events = []
        packets = internal_buffer.split(self._stop_word)

        complete_event = self._timeout_event(packets)
        if complete_event:
            complete_events.append(complete_event)

        # Separate the packets into events and answers
        c_events, answers = self._sort_packets(packets[:-1])
        complete_events.extend(c_events)

        # Create raw events from the complete_events
        for complete in complete_events:
            raw_event = self._create_raw_event(complete)
            raw_events.append(raw_event)

        return raw_events, answers, packets[-1]

    def _timeout_event(self, packets) -> dict:
        """Check if an incomplete event has timed out.

        If the event has timed out, it is added to the list of completed events.
        This means if the data stops, the incomplete event is not lost.

        Args:
            packets (list): list of packets to check for timeout.

        Returns:
            The completed event.
        """
        complete_event = None
        if (
            self._do_timeout and len(packets) > 1
        ):  # Not sure if there need to be a packet received for the timeout to activate.
            if time.perf_counter() - self._last_packet_time > self._timeout:
                if (
                    self._incomplete_event["timing"] != UNUSED_TIMING
                ):  # If the incomplete event is not new.
                    complete_event = self._incomplete_event
                self._create_incomplete_event(unused_timing=UNUSED_TIMING)
                logger.debug(f"Packager timed out, flushed an incomplete events")
            self._last_packet_time = time.perf_counter()

        return complete_event

    def _create_incomplete_event(self, unused_timing: int):
        """Create an incomplete event."""
        self._incomplete_event = defaultdict(list)
        self._incomplete_event["timing"] = unused_timing

    def _sort_packets(self, packets: List[bytes]) -> list:
        """Sorts packets received into answers and incomplete events.

        Sorts the packets into answers and incomplete events based on the length.
        An answer from the board is 8 bytes long, and an incomplete event is 72 bytes long.

        Args:
            packets (list): the packets to sort

        Returns:
            A list of events and a list of answers from the input packets.
        """
        events = []
        answers = []
        for packet in packets:
            if self._is_answer_package(packet):
                answer_package = self._create_answer_package(packet + self.stop_word)
                answers.append(answer_package)
            elif self._is_data_package(packet):
                event = self._insert_packet(
                    packet
                )  # Event is None if no complete event
                if event:
                    events.append(event)

        return events, answers

    def _create_raw_event(self, event: dict) -> dict:
        """Generates a raw event dict for each complete event
        given.

        Args:
            complete_events (list): list of complete events.
                Each element should be a `dict[chan: list[bytes]]`

        Returns:
            A list of parse-able raw events.
        """
        stop_word = self._stop_word
        all_packets = [
            x for chan in event if isinstance(chan, int) for x in event[chan]
        ]
        output_event = self._create_data_package(
            rawdata=stop_word.join(all_packets) + stop_word, pkg_num=self._pkg_num
        )
        self._pkg_num += 1
        return output_event

    def _is_data_package(self, package: bytearray) -> bool:
        """Checks if a package is a data (event) package based on length.

        Args:
            package (bytearray): the raw package.

        Returns:
            True if the package is data, False otherwise.
        """
        return len(package) == self._package_size

    def _insert_packet(self, packet: bytearray):
        """Sorts a package into the earliest occuring incomplete event for which there is a
        slot available. If the event is complete, it's returned

        If all previous events have the correct number of windows for that channel, a new
        incomplete event is created and the package is stored into it.

        Args:
            packet (bytearray): the package to store
        """
        complete_event = None
        word_builder = (
            lambda i: (packet[i] << 8) + packet[i + 1]
        )  # Builds a word from a pair of bytes
        packet_channel = word_builder(0) & 0b111111

        # packet timing are the lower 12 bit of the second and third word, making a 24-bit timing.
        packet_timing = (
            int(word_builder(2) & self._timing_mask) << self._timing_shift
        ) | int(word_builder(4) & self._timing_mask)

        # Case 1: timing withing margins, it's part of THIS event
        if (
            self._incomplete_event["timing"] - self.timing_margin
            <= packet_timing
            <= self._incomplete_event["timing"] + self.timing_margin
        ):
            self._incomplete_event[packet_channel].append(packet)  # NEED to append

        # Case 2: timing outside margins, it's part of a NEW event
        elif (
            self._incomplete_event["timing"] + self.timing_margin <= packet_timing
            or packet_timing < self._incomplete_event["timing"] - self.timing_margin
        ):
            if (
                self._incomplete_event["timing"] != UNUSED_TIMING
            ):  # The first packet is never complete
                complete_event = self._incomplete_event

            self._create_incomplete_event(packet_timing)
            self._incomplete_event[packet_channel].append(packet)

        else:
            pass

        return complete_event

    def stop(self):
        super().stop()
        # Any incomplete data will be saved in case stop is in middle of event.
        if self._incomplete_event["timing"] != UNUSED_TIMING:
            complete = self._incomplete_event
            raw_event = self._create_raw_event(complete)
            self.output_buffer.append(raw_event)



"""Connection factory.

Creates and returns a connection of the specified type.
"""
from logging import getLogger

from ._MockUART import MockUART
from ._UART import UART
from .tcp import TCP
from .udp import UDP

LOGGER = getLogger(__name__)

# This list gets updated based on whether or not certain
# connection types are actually available
VALID_CONNECTION_TYPES = [
    "ftdi",
    "tcp",
    "udp",
    "uart",
    "ft60x",
    "none",
]
try:
    from ._FTDI import FTDI
except (ImportError, FileNotFoundError) as e:
    VALID_CONNECTION_TYPES.remove("ftdi")
    LOGGER.warning("FTDI import failed: %s", e)
try:
    from ._USB import USB3
except (ImportError, FileNotFoundError) as e:
    VALID_CONNECTION_TYPES.remove("ft60x")
    LOGGER.warning("D3XX import failed: %s", e)

FTDI_AVAILABLE = "ftdi" in VALID_CONNECTION_TYPES
USB3_AVAILABLE = "ft60x" in VALID_CONNECTION_TYPES


def get_connection(conn_info):
    """Setup a connection to the device and raise the connection flag.

    Takes a connection_info dictionary and sets the connection accordingly.

    Input:
        conn_info (dict):
            "type": (str) "none" "eth" "uart"
            "usb_port": (str) port name
            "speed": (int) baud rate or ethernet speed
            "ip_addr": (str) should be in XXX.XXX.XXX.XXX format
            "ip_port": (int) port number

    Raises:
        TypeError if the connection type is unknown.
        ConnectionError if connection can't be created.
    """
    conn_type = _get_connection_type_or_raise(conn_info)
    connection = None
    try:
        if conn_type == "tcp":
            connection = _set_tcp_connection(conn_info)
        elif conn_type == "udp":
            connection = _set_udp_connection(conn_info)
        elif conn_type == "uart":
            connection = _set_uart_connection(conn_info=conn_info)
        elif conn_type == "ftdi":
            connection = _set_ftdi_connection(conn_info=conn_info)
        elif conn_type == "ft60x":
            connection = _set_ft60x_connection(conn_info=conn_info)
    except (ConnectionError, Exception) as error_msg:
        raise ConnectionError(error_msg)

    return connection


def _set_uart_connection(conn_info):
    """Connect to the UART.

    Setup the connection to serial.
    sets self.connection to the UART connection.

    Input:
        serial_port (str): com-port(win) or device address(*nix)
        baud (int): Set communication baud rate.
    """
    baud = conn_info["speed"]
    connection = None
    LOGGER.info("connecting using %s, %s", conn_info, baud)

    try:
        if conn_info.get("mock", False):
            connection = MockUART(conn_info)
        else:
            connection = UART(conn_info)
    except ConnectionError:
        LOGGER.debug("Couldn't open UART.UART(%s, %s)", conn_info, baud)
        raise
    else:
        connection.open()
        connection.toggleLoopback(0)
    return connection


def _set_ftdi_connection(conn_info):
    """Connect using the FTDI driver.

    The FTDI connection uses device number rather than comport.
    Allowing connections to devices not enuemrated.

    """
    if not FTDI_AVAILABLE:
        raise ConnectionError("Can't use FTDI connections if driver is unavailable.")
    baud = conn_info["speed"]
    connection = None
    LOGGER.info("connecting using %s, %s", conn_info, baud)

    try:
        if conn_info.get("mock", False):
            connection = MockUART(conn_info)
        else:
            connection = FTDI(conn_info)
    except ConnectionError:
        LOGGER.debug("Couldn't open FTDI(%s, %s)", conn_info, baud)
        raise
    else:
        connection.open()
    return connection


def _set_ft60x_connection(conn_info):
    """Connect using the FTDI driver.

    The FTDI connection uses device number rather than comport.
    Allowing connections to devices not enuemrated.

    """
    if not USB3_AVAILABLE:
        raise ConnectionError("Can't use USB3 connections if driver is unavailable.")
    connection = None
    LOGGER.info("connecting using %s", conn_info)

    try:
        if conn_info.get("mock", False):
            raise ConnectionError("Mock connection is not available for USB3.")
        connection = USB3(conn_info)
    except ConnectionError:
        LOGGER.debug("Couldn't open USB3(%s)", conn_info)
        raise
    else:
        connection.open()
    return connection


def _set_tcp_connection(conn_info):
    """Connect to the UART.

    Setup the connection to serial.
    sets self.connection to the UART connection.

    Input:
        serial_port (str):
        baud (int): Set communication baud rate.

    """
    try:
        ip = conn_info["ip"]
        port = conn_info["port"]
        stop_word = conn_info.get("stop_word", b"\xFA\xCE")
    except KeyError as e:
        raise ValueError(f"Missing required parameter: {e}")

    connection = None
    try:
        connection = TCP(ip, port, stop_word)
        connection.open()
    except ConnectionError:
        raise
    return connection


def _set_udp_connection(conn_info):
    """Connect to over UDP socket.

    sets self.connection to the UDP connection.

    Args:
        conn_info (dict): containing the connection params for the UDP connection obj.

    Returns:
        An open UDP connection
    """
    try:
        conn_info["stop_word"]
    except KeyError as e:
        raise ValueError(f"Connection is missing required parameter 'stop_word'")

    connection = None
    try:
        connection = UDP(conn_info)
        connection.open()
    except ConnectionError:
        raise
    return connection


def _get_connection_type_or_raise(conn_info: dict):
    conn_type = conn_info.get("type", None)
    if conn_type is None:
        raise AttributeError('Connection info missing "type" key')
    if not isinstance(conn_type, str):
        raise TypeError("Connection type must be a string")
    conn_type = conn_type.lower()
    if conn_type not in VALID_CONNECTION_TYPES:
        raise TypeError(f"{conn_type} is not a recognized connection type.")
    return conn_type



import logging

from naludaq.communication import DigitalRegisters
from naludaq.controllers import get_readout_controller
from naludaq.tools.threshold_scan.hdsoc_thresholdscan import ThresholdScanHdsoc

import numpy as np
import time

LOGGER = logging.getLogger("naludaq.threshold_scan.hdsocv2_eval")


BROADCAST = 64


class ThresholdScanHdsocv2(ThresholdScanHdsoc):
    """Tool for sweeping over trigger thresholds and counting number of triggers.
    Can be used to determine the ideal point to set the trigger.

    """

    def __init__(
        self,
        board,
        *args,
        **kwargs,
    ):
        """Threshold scan for hdsoc boards.

        Args:
            board (Board): board object
            start_value (int): the start value for the scan
            stop_value (int): the stop value for the scan
            step_size (int): the step size for the scan
            low_ref_value (int): low reference value
            high_ref_value (int): high reference value
        """
        super().__init__(board, *args, **kwargs)
        self._references = {
            "0_15": self.references,
            "16_31": self.references,
            "32_47": self.references,
            "48_63": self.references,
        }

    def _get_scalar_values(self, pause: float = 0.1):
        """Reimplementation to perform some extra setup."""
        get_readout_controller(self.board).set_readout_channels(self.channels)
        tc = self.board.trigger
        tc.references = self._references
        tc.tsel = True
        return self._get_scalar_values_vertical(pause)

    def _get_scalar_values_vertical(self, pause: float) -> np.ndarray:
        """Scan the range and return np.array with trigger amounts

        Args:
            pause (float): amount of seconds to pause in between samples.

        Returns:
            Triggers value per channel as `np.array`.
        """
        scan_values = self.scan_values
        output = np.zeros((self.board.channels, len(scan_values)))

        # NOTE: the default threshold scan iterates over scan values then channels,
        # but on HDSoC it is super important that this order is reversed!
        # Otherwise the threshold scan will be super messed up or not work at all.
        for ch_i, chan in enumerate(self.channels):
            for i, value in enumerate(scan_values):
                if self._cancel:
                    break
                progress = ch_i * len(scan_values) + i + 1
                progress_total = len(self.channels) * len(scan_values)
                LOGGER.debug("Scan progress: %s/%s", progress, progress_total)
                self._progress.append(
                    (
                        int(95 * progress / progress_total),
                        f"Scanning {progress}/{progress_total}",
                    )
                )
                default_tval = self.board.trigger.params.get("default_trigger_value", 0)
                tvalues = {ch: default_tval for ch in range(self.board.channels)}
                tvalues[chan] = int(value)
                self.board.trigger.values = tvalues
                time.sleep(pause)

                scalar = self._get_single_scalar_value(chan)
                output[chan][i] = scalar

        return output

    def _backup_board_settings(self):
        """Back up the trigger values, readout channels, references, and tsel"""
        self._original_trigger_values = self.board.trigger.values.copy()
        rc = get_readout_controller(self.board)
        self._original_readout_channels = rc.get_readout_channels()
        self._original_references = self.board.trigger.references

    def _restore_board_settings(self):
        """Restore the trigger values, readout channels, references, and tsel"""
        self.board.trigger.values = self._original_trigger_values
        self.board.trigger.references = self._original_references

        rc = get_readout_controller(self.board)
        rc.set_readout_channels(self._original_readout_channels)

        dr = DigitalRegisters(self.board)
        dr.write("selectchannel", BROADCAST)

    def _set_tsel_enabled(self, left: bool, right: bool):
        """Set whether tsel is enabled on each side.

        During the threshold scan, tsel must be enabled for scalars
        to read correctly.
        """
        pass
        # self._trigger_controller._set_tsel_enabled()



from naludaq.communication import AnalogRegisters

from .default import TriggerController


class TriggerControllerAodsoc(TriggerController):
    """Trigger controller for AODSOC boards."""

    def __init__(self, board):
        super().__init__(board)

        self._num_chips = board.params.get("num_chips", 2)
        self._num_trig_chans_per_chip = board.params.get("num_trig_chans_per_chip", 4)

    def _set_trigger_thresholds(
        self, trigger_values: "dict[int, int] | int | list[int] | None" = None
    ):
        """Sets the trigger values to those in the board object and writes to hardware."""
        values = trigger_values or self.values
        self._write_analog_registers_per_chip("trigger_threshold_{:02}", values)

    def _set_wbiases(self):
        """Set the wbias registers based on the board trigger values"""
        wbias = {
            ch: self.wbias[ch] if bool(val) else 0 for ch, val in self.values.items()
        }
        self._write_analog_registers_per_chip("wbias_{:02}", wbias)

    def _set_trigger_offsets(self):
        """Set the offset registers to those in the trigger values"""
        vofs = self.offsets
        self._write_analog_registers_per_chip("offset_{:02}", vofs)

    def _write_analog_registers_per_chip(
        self, reg_name_format: str, values: "list | dict"
    ):
        """Write analog registers with a given register name format using the
        values provided. Writes values to both chips individually.

        Args:
            reg_name_format (str): register name format string, must take one integer argument.
            values (list|dict): list or dict of values to write. Should be all 8 values across both chips.
        """
        for chip in range(self.board.available_chips):
            for idx in range(self._num_trig_chans_per_chip):
                name = reg_name_format.format(idx)
                val = values[idx + chip * self._num_trig_chans_per_chip]
                AnalogRegisters(self.board, [chip]).write(name, val)



"""Base class for registers.
"""
import abc
from abc import ABC
from collections import defaultdict
from functools import wraps
from logging import getLogger
from typing import Iterable, List

from naludaq.backend.exceptions import ConnectionError
from naludaq.backend.managers.connection import ConnectionManager
from naludaq.backend.managers.io import BoardIoManager
from naludaq.helpers.exceptions import InvalidRegisterError

from . import _common as helpers

LOGGER = getLogger(__name__)


def clear_buffer(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """Clears the board buffer before"""
        if self.board.using_new_backend:
            device = ConnectionManager(self.board).device
            if device is None:
                raise ConnectionError("There is no active connection")
            device.clear_buffers()
        else:
            self.board.connection.reset_input_buffer()
        rv = func(self, *args, **kwargs)

        return rv

    return wrapper


class Registers(ABC):
    """Base class for communication with registers.
    This class should not be used directly; use one of the subclasses instead.

    Inheritence tree:

    ```mermaid
    graph TD;
        Registers --> ChipRegisters
        Registers --> FpgaRegisters
        ChipRegisters --> DigitalRegisters
        ChipRegisters --> AnalogRegisters
        FpgaRegisters --> ControlRegisters
        FpgaRegisters --> I2CRegisters
    ```

    Atrributes:
    -----------

        board: The board to update the registers on.
        width_cmd (int): width of the command type in hex chars
        width_addr (int): width of the address in hex chars
        width_val (int): width of the value in hex chars
        width_total (int): total width of the command in hex chars

    """

    def __init__(self, board, register_type: str):
        self.board = board
        self._register_type = register_type
        if not isinstance(getattr(board, "registers", None), dict):
            raise InvalidRegisterError("The board is lacking valid registers")
        if not board.registers.get(register_type, None):
            raise InvalidRegisterError(f"The registers is lacking {register_type}")
        self._width_cmd = 3
        self._width_addr = 2  # hex char
        self._width_val = 3  # hex char
        self._width_total = 8  # common firmware

        self._stopword = None
        self.stopword = self.board.params.get(
            "register_stop_word",
            self.board.params["stop_word"],
        )

    @property
    def stopword(self):
        """Get/Set stopword"""
        return self._stopword

    @stopword.setter
    def stopword(self, value):
        if not isinstance(value, (int, str, bytes)):
            raise TypeError(f"stopword must be str, bytes or hex, got {type(value)}")
        elif isinstance(value, str):
            if len(value) % 2 != 0:
                raise ValueError(
                    "Since stopword converts to bytes, it must be of even length."
                )
            value = bytes.fromhex(value)
        elif isinstance(value, int):
            value = bytes.fromhex(f"{value:X}")
        self._stopword = value

    @property
    def registers(self) -> dict:
        """Shortcut to board registers"""
        return self.board.registers[self._register_type]

    @abc.abstractmethod
    def set(self, name: str, value: int):
        """Set the value of a register without writing to the board.

        Args:
            name (str): Name of the register.
            value (int): Value to set the register to.
        """

    @abc.abstractmethod
    def set_addr(self, addr: "str | int", value: int):
        """Set the value of registers on the given address without
        writing to the board.

        Args:
            addr (str | int): Address of the register.
            value (int): Value to set the register to.
        """

    # VALIDATIONS ############################################################
    #
    ##########################################################################
    def _validate_addr_or_raise(self, addr: str):
        """Validate the address type and make sure it exists.

        Args:
            addr(str): Hex value of address.

        Raises:
            ValueError if addr is not a valid hex or if it doesn't exist in register
            TypeError if addr is not a string or int
        """
        if not isinstance(addr, str):
            raise TypeError("Register number needs to be an integer")
        if addr.zfill(self._width_addr).upper() not in self.list_addresses():
            raise ValueError(f"Address {addr} does not exist in {self._register_type}")

    def _validate_name_or_raise(self, name):
        """Name is a user input, make sure it's valid.

        Args:
            name(str): Name of the analog register.

        Returns:
            True if name is valid, False if it's not.

        Raises:
            TypeError and ValueError
        """
        if not isinstance(name, str):
            raise TypeError(f"Name: {name} is of the wrong type: {type(name)}")
        if name.casefold() not in [k.casefold() for k in self.registers.keys()]:
            raise ValueError(f"Name: {name} is not a valid analog register.")

    def _validate_names_or_raise(self, names: Iterable[str]):
        """Validate list of register names or raise an error."""
        if not isinstance(names, Iterable):
            raise TypeError("Names must be Iterable[str]")
        for name in names:
            self._validate_name_or_raise(name)

    def _validate_value_or_raise(self, value: int, bitwidth: int):
        """Value is a user input, make sure it's valid.

        Args:
            value: Value to set the analog register to.

        Raises:
            TypeError: if the value is of the wrong type.
            ValueError: if the Value is not positive.
        """
        if not isinstance(value, (type(None), int)):
            raise TypeError(f"Value argument is of the wrong type: {type(value)}")
        if isinstance(value, int) and not 0 <= value < 2**bitwidth:
            raise ValueError("Value argument must be  a positive integer.")

    def _validate_value_map_or_raise(self, values: dict[str, int]):
        """Validate the given value map or raise an error"""
        if not isinstance(values, dict):
            raise TypeError("Values must be dict[str, int]")
        for name, value in values.items():
            self._validate_name_or_raise(name)
            bitwidth = self.registers[name]["bitwidth"]
            self._validate_value_or_raise(value, bitwidth)

    # HELPERS ################################################################
    #
    # A set of helper functions used by all registers
    #
    ##########################################################################

    def _get_addr_from_name_or_raise(self, name):
        """Returns the control registers addr from the name.

        Args:
            name(str): Name of the control register

        Returns:
            The number of the control register
        """
        if not isinstance(name, str):
            raise TypeError("Register name must be a strign")

        addr = self.registers.get(name.casefold(), False)
        if addr is False:
            raise InvalidRegisterError(f"{name} is not a valid register.")
        addr = addr["address"]
        if isinstance(addr, int):
            addr = f"{addr:X}"  # pragma: no cover
        addr = addr.zfill(self._width_addr)
        return addr.upper()

    def _validate_command_or_raise(self, command: str):
        """Make sure a generated command is of the right length."""
        if len(command) != self.width_total:
            raise ValueError(
                f"Generated command is invalid. A command must be {self.width_total} char, got {command} or length {len(command)}"
            )  # pragma: no cover

    # COMMON #################################################################
    #
    # A list of convenience tools, common for all registers.
    #
    ##########################################################################
    def list(self) -> list:
        """Lists the name of all available registers.

        Returns:
            A list of all register names
        """
        return list(k.lower() for k in self.registers.keys())

    def list_registers(self) -> dict:
        """Returns all available registers as a dict, organized in the
        same way as the registers file.

        Returns:
            A dict containing all registers
        """
        return self.registers

    def list_addresses(self) -> List[str]:
        """Generate a list of all available addresses.

        This is useful since an address can contain multiple registers.
        The register name can correspond to one or many bits at a specific address.

        Returns:
            List of available addresses in hex (str)
        """
        reg_addr = {
            helpers.normalize_address(value["address"], self._width_addr)
            for value in self.registers.values()
        }
        return sorted(list(reg_addr), key=lambda x: int(x, 16))

    def registers_on_address(self, addr: str) -> dict[str, dict]:
        self._validate_addr_or_raise(addr)
        addr = helpers.normalize_address(addr, self._width_addr)
        return {
            name: value
            for name, value in self.registers.items()
            if helpers.normalize_address(value["address"], self._width_addr) == addr
        }

    def show(self, name: str) -> str:
        """Returns the internal contents of a register.

        Args:
            name (str): The name of the register

        Returns:
            A string representation of the given register.

        Raises:
            TypeError if the name is not a string
            ValueError if the name is not a valid register
        """
        self._validate_name_or_raise(name)

        return self.registers[name]

    def _send_command(self, cmd):
        """Send request to the board"""
        if self.board.using_new_backend:
            BoardIoManager(self.board).write(cmd)
        else:
            self.board.connection.send(cmd)

    def __repr__(self):
        return f"{type(self).__name__}: {self.registers}(board = {self.board})"  # pragma: no cover


class WriteRegistersABC(Registers):
    def write(self, name: str, value: int = None):
        if value is not None:
            self.set(name, value)
        command = self.generate_write(name, value)
        self._send_command(command)

    def write_addr(self, addr: "int | str", value: int = None):
        if value is not None:
            self.set_addr(addr, value)
        command = self.generate_write_addr(addr, value)
        self._send_command(command)

    def write_all(self):
        for addr in self.list_addresses():
            self.write_addr(addr)

    def write_many(self, value_map: dict[str, int]):
        """Write several registers at once.

        Args:
            registers (dict[str, int]): mapping of register values to names
        """
        self._validate_value_map_or_raise(value_map)

        # update all software registers first to prevent generate_write
        # from using potentially old register values in subsequent writes
        for name, value in value_map.items():
            self.set(name, value)

        addresses = {
            self._get_addr_from_name_or_raise(name) for name in value_map.keys()
        }
        commands = [self.generate_write_addr(addr) for addr in addresses]
        command = "".join(commands)
        self._send_command(command)

    @abc.abstractmethod
    def generate_write(self, name: str, value: int = None) -> str:
        pass

    @abc.abstractmethod
    def generate_write_addr(self, addr: str, value: int = None) -> str:
        pass

    def _generate_write_atom(self, addr: str, value: int) -> str:
        cmd = f"{self._write_cmd:X}"
        cmd_addr = addr.zfill(self._width_addr)
        val = f"{value:X}".zfill(self._width_val)
        command = f"{cmd}{cmd_addr}{val}"
        return command.upper()


class ReadRegistersABC(Registers):
    def read(self, name: str) -> dict:
        """Read from a register with a name.

        Args:
            name: Name of the register.

        Raises:
            ValueError if name is not in register.
            TypeError if name is not a str.
        """
        self._validate_name_or_raise(name)
        addr = self.registers[name]["address"]
        return self.read_addr_named(addr)[name]

    def read_addr(self, addr: "int | str") -> int:
        """Read an address and return the value.

        Args:
            addr (str | int): 8-bit address to read as either int or hex.
        """
        command = self.generate_read_addr(addr)
        return self._read_response(command)

    def read_addr_named(self, addr: "int | str") -> dict:
        """Read an address and return the register dicts for all registers
        on that address.

        Args:
            addr (str | int): address to read as either int or hex.

        Returns:
            dict: {name: register_dict}
        """
        addr_value = self.read_addr(addr)
        return {
            name: reg.copy() | {"value": helpers.full_to_partial_value(addr_value, reg)}
            for name, reg in self.registers.items()
            if reg["address"].upper() == addr.upper()
        }

    def read_all(self) -> list:
        """Read all digital registers from the board.

        Reads the values based on the register number rather than the name.

        Args:
            board

        Returns:
            dictionary with {reg_num: values,}
        """
        if self.board.using_new_backend:
            addresses = self.list_addresses()
            commands = [self.generate_read_addr(addr) for addr in addresses]
            values = {}
            answers = BoardIoManager(self.board).read_all(commands)
            for address, answer in zip(addresses, answers):
                parsed_answer = self._parse_response(answer)["value"]
                reg = self._registers_from_addr_value(address, parsed_answer)
                values.update(reg)
        else:
            values = {
                name: register
                for addr in self.list_addresses()
                for name, register in self.read_addr_named(addr).items()
            }
        return values

    def read_many(self, names: list[str]) -> dict:
        """Read several registers at once.

        Do not use this function to read the same register multiple times,
        as older values will be discarded.

        Args:
            names (Iterable[str]): names of registers to read

        Returns:
            list[int]: list of register values in the same order as provided.

        Raises:
            TimeoutError: if one or more registers could not be read in time.
        """
        self._validate_names_or_raise(names)

        # cannot perform concurrent reads without naludaq_rs
        if not self.board.using_new_backend:
            return [self.read(name)["value"] for name in names]

        # sort registers & generate by address to reduce # of commands
        addresses = defaultdict(list)
        for name in names:
            addr = self._get_addr_from_name_or_raise(name)
            addresses[addr].append(name)
        read_cmds = [self.generate_read_addr(addr) for addr in addresses.keys()]

        # response order is guaranteed to be the same as the commands sent
        try:
            responses = BoardIoManager(self.board).read_all(read_cmds)
        except TimeoutError:
            raise

        # parse responses into registers by address
        values = {}
        for (addr, addr_names), response in zip(addresses.items(), responses):
            addr_value = self._parse_response(response)["value"]
            registers = self._registers_from_addr_value(addr, addr_value)
            for name in addr_names:
                values[name] = registers[name]["value"]

        sorted_values = [values[name] for name in names]
        return sorted_values

    def generate_read(self, name: str) -> str:
        """Generate and return the read command as hex"""
        return self.generate_read_addr(self.registers[name]["address"])

    @abc.abstractmethod
    def generate_read_addr(self, addr: "int | str") -> str:
        """Generate and return the read addr command as hex"""
        pass

    def _generate_read_atom(self, addr: "int | str") -> str:
        """Generate a read command for a single address.

        Args:
            addr (int | str): Address to read from.

        Returns:
            str: Command to send to the board.
        """
        if isinstance(addr, int):
            addr = f"{addr:X}"
        cmd = f"{self._read_cmd:X}"
        cmd_addr = addr.zfill(self._width_addr)
        val = "0".zfill(self._width_val)
        command = f"{cmd}{cmd_addr}{val}"
        return command.upper()

    @clear_buffer
    def _read_response(self, command: str) -> int:
        """Send a read command to the board and parse the response.

        Args:
            command (str): command to send to the board.

        Returns:
            int: value read from the board. Value is -1 if the read failed.
        """
        try:
            if self.board.using_new_backend:
                response = BoardIoManager(self.board).read(command)
            else:
                self._send_command(command)
                response = self.board.connection.read_until(self.stopword)
        except TimeoutError:  # pragma: no cover
            return -1  # pragma: no cover
        return self._parse_response(response)["value"]

    def _parse_response(self, buffer: bytes) -> dict:
        """Parses a raw binary answer into a dictionary.

        Args:
            buffer (bytes): Response to parse.

        Returns:
            dict: parsed response containing keys: "header", "read_reg", "value".
        """
        header = int.from_bytes(buffer[0:2], byteorder="big", signed=False)
        value = int.from_bytes(buffer[2:4], byteorder="big", signed=False)
        return {
            "header": header,
            "read_reg": header & 0xFF,
            "value": value,
        }

    def _registers_from_addr_value(self, addr: str, addr_value: int) -> dict:
        _registers = {}
        for name, reg in self.registers.items():
            if reg["address"].upper() == addr.upper():
                _registers[name] = reg.copy()
                _registers[name]["value"] = helpers.full_to_partial_value(
                    addr_value, reg
                )

        return _registers

        return {
            name: {"value": helpers.full_to_partial_value(addr_value, reg)}
            for name, reg in self.registers.items()
            if reg["address"].upper() == addr.upper()
        }


# validation #################################################################
#
##############################################################################
def validate_registermap_or_raise(registers):
    """Validate a regmap, raises error if it's not valid."""

    # invert registers
    inv_register = invert_regmap(registers)

    # find potential conflicts
    try:
        conflicts = find_conflicting_registers(inv_register)
    except Exception as e_msg:
        raise InvalidRegisterError(f"Following register have conflicting bits: {e_msg}")
    # if potential conflicts
    if conflicts:
        raise InvalidRegisterError(
            f"Following register have conflicting bits: {conflicts}"
        )

    return True


def find_conflicting_registers(registers):
    """Takes an inverted regmap and finds all conflicts

    Args:
        registers (dict): register dictionary and find all conflicts.

    Returns:
        dictionary with conflictingaddresses and conflicting names.
    """
    conflict_regs = defaultdict(list)
    inv_regmap = invert_regmap(registers)  # {addr: [(name, reg)]}
    for addr, regs in inv_regmap.items():
        if len(regs) <= 1:
            continue

        regs = sorted(regs, key=lambda i: i[1]["bitposition"])
        for idx, (name1, item1) in enumerate(regs[:-1]):
            item1_end = item1["bitposition"] + item1["bitwidth"] - 1

            for name2, item2 in regs[idx + 1 :]:
                if item2["bitposition"] > item1_end:
                    break
                conflict_regs[addr].append(name1)
                conflict_regs[addr].append(name2)

    conflict_regs = {addr: sorted(set(names)) for addr, names in conflict_regs.items()}
    return conflict_regs


def invert_regmap(registers):
    """Invert a regmap dict to return addr: register instead of name: register

    Args:
        registers (dict): register map {name: registers}

    Returns:
        Dictionary with {address: register}
    """
    all_regs = defaultdict(list)

    for key, val in registers.items():
        addr = val["address"]
        all_regs[addr].append((key, val))

    return all_regs



""" Board controller for the ASoCv3s

This board has a few differences, the stop acq needs to send a proper stop acq command.

"""
import time
from logging import getLogger

from naludaq.communication.control_registers import ControlRegisters

from .default import BoardController

LOGGER = getLogger("naludaq.board_controller_asocv3s")


class ASoCv3SBoardController(BoardController):
    def start_readout(
        self,
        trig: str = "self",
        lb: str = "trigrel",
        acq: str = "raw",
        dig_head: bool = False,
        ped: str = "zero",
        readoutEn: bool = True,
        singleEv: bool = False,
    ):
        """Sends a readout signal to the board.

        Obstensibly reads out the chip?

        Args:
            trigger_type (str):
                'immediate', '00' - digitize whatever instantly as fast as possible.
                'ext': '01' - use the external trigger.
                'self': '10' - trigger on input RF signal.
            lbType (str):
                'forced': '00' - always reads out same set of windows, start window set by lookback
                'trig': '01' - relative to the trigger.
                'roi': '10' - complicated...
            acq_type (str):
                'raw': '00' - Can't ped-subtract. <Default output>
                'ped': '01' - sub, spits out headers, channels, and windows. <Requires firmware modification>
            dig_head (bool):
                True: reads out a "digitization" header after each sram read (only works with acq_type="ped")
                False: doesn't read out that header
            pedMode (str):
                'z': '00' - all peds = 0
                'c': '01' - peds = chan*1024, set offset per channel
                'r': '11' - peds = RAM addr
            readoutEn (bool): Enables readout to ethernet (otherwise data thrown away)
                self.readReadoutRate() is the companion to this to see the true rate
                readout rate scalar stored in FPGA reg N_GPR(32) + 5
            singleEV (bool): True - only reads one event.

        Command
        -------

        command is: BMMXXVVV
        MM = singleEvent(7) + readoutEn(6) + pedSel(5 dt 4) + modeSel(3 dt 0)
        - modeSel=x"0" analog write
        - modeSel=x"1" digital write
        - modeSel=x"2" digital read
        - modeSel=x"3" waveform read

        VVVVV = addr_sel(19 dt 12) + data_sel(11 dt 0)
        - data_sel(5 downto 4) = trig type
        - data_sel(3 downto 2) = lb type
        - data_sel(1 downto 0) = acq type

        """
        self.serial_to_parallel_conv(True)

        super().start_readout(
            trig,
            lb,
            acq,
            dig_head,
            ped,
            readoutEn,
            singleEv,
        )

    def serial_to_parallel_conv(self, enabled: bool):
        """Enabled/Disable serial to parallel conversion

        This conversion is used to enabled a 16-bit padded output
        instead of the 12-bit raw output.

        The parser requires this to be enabled.
        Register reads require this to be disabled.
        """

        ControlRegisters(self.board).write("ser2par_en", enabled)

    def stop_readout(self):
        """Toggles the "stopacq" signal on the readout module.

        It's equivalent of asking it nicely to stop reading.
        """
        self.is_reading_out = False
        cmd = "B0B00000"
        self._send_command(cmd)
        time.sleep(0.1)
        self.serial_to_parallel_conv(False)

    def set_loopback_enabled(self, enabled: bool):
        """Set whether serial loopback is enabled.

        Loopback can safely be disabled during most of the operations with the board.
        Loopback **must** be disabled when communicating over the serial interface.
        If serial communication with the ASIC is intended then this should run during startup and only be enabled as needed.

        Args:
            enabled (bool): True to enable loopback.

        Raises:
            TypeError if enabled is not a bool.
        """
        if not isinstance(enabled, bool):
            raise TypeError("Argument must be bool")
        OFF = "B0900002"
        ON = "B0900003"
        cmd = ON if enabled else OFF
        self._send_command(cmd)



import abc


class HeaderParser(abc.ABC):
    """Base class for header parsers.

    All subclasses must implement the static `parse_header` method.
    """

    @staticmethod
    @abc.abstractmethod
    def parse_header(event: dict, raw_data: bytes):
        """Parse event headers from raw data and store them into an event dict.

        Args:
            event (dict): the event
            raw_data (bytes): the raw data to parse.
        """


class NoOpHeaderParser(HeaderParser):
    """No-op header parser, doesn't do anything. Used only as a placeholder
    for data that doesn't have event headers.
    """

    @staticmethod
    def parse_header(event: dict, raw_data: bytes):
        """Placeholder for

        Args:
            event (dict): the event
            raw_data (bytes): the raw data to parse.
        """



import logging
from collections import deque
from functools import partial

import numpy as np

from naludaq.board import Board
from naludaq.controllers import get_dac_controller
from naludaq.helpers.exceptions import (
    BadDataError,
    InvalidOptimizationParameters,
    IterationError,
    OperationCanceledError,
    OptimizationError,
)
from naludaq.helpers.helper_functions import (
    find_missing_channels,
    group_channels_by_chip,
)
from naludaq.tools.data_collector import get_data_collector

from .channel_writer import UDC16ChannelWriter

LOGGER = logging.getLogger("naludaq.ConversionRampOptimizer")


class UDC16ConversionRampOptimizer:
    """Generates tuning parameters for isel ramp current and cap select
    Tuning parameters are of the form:
    {
        0 (Channel): {
            isel_ramp_current: (value),
            isel_cap_select: (value),
        }
    }
    """

    def __init__(self, board: Board):
        self.board = board

        self._progress = []
        self._canceled = False
        self._num_channels = board.channels
        self._num_chips = board.available_chips
        self._channels_per_chip = self._num_channels // self._num_chips
        self.target_value = 500

        self.channel_writer = UDC16ChannelWriter(self.board)
        self.channel_writer.reset_mask_after_write = False
        self._last_event_channels = []
        self._bad_channels = []

    @property
    def progress(self):
        """Get/Set the progress message queue.

        This is a hook the read the progress if running threads.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    def run(self, channels: list[int]):
        """Tunes the isel ramp and cap to put the given channels into the midrange.
        Writes the parameters per channel and stores the tuning values to board.tuning.

        Args:
            channels (list[int]): List of channels to tune

        Raises:
            OptimizationError: Something went wrong with running the optimizing script
            TimeoutError: Data capture timedout/couldn't get data
            OperationCanceledError: The operation was canceled by the user
        """
        suggested_parameters = self.generate(channels)
        self.validate_iselrampcap_values(suggested_parameters)
        if suggested_parameters:
            # Status message ###############################
            _update_progress(
                self.progress,
                80,
                f"Calibrating ramp:",
            )

            self._update_isel_for_channels(suggested_parameters)
            self.board.tuning["conversion_ramp"] = suggested_parameters
        else:
            raise OptimizationError(
                "Optimization for Conversion Ramp failed. Unable to create optimization parameters"
            )

        if self._bad_channels:
            raise OptimizationError(self._chan_data_err_str(channels))

    def generate(self, channels: list[int]):
        """Determines the isel ramp and cap to put the average values of the given
        channels to the target value (default=500). Uses linear regression to
        determine what values to set isel ramp current and cap select to put the
        average channel value to the target value.

        Args:
            channels (list[int]): List of channels to tune

        Returns:
            suggested_parameters (dict): Dictionary where it maps the
                register, to a list of values per channel.
            Tuning parameters are of the form:
            {
                0 (Channel): {
                    isel_ramp_current: (value),
                    isel_cap_select: (value),
                }
            }

        Raises:
            OptimizationError: Something went wrong with running the optimizing script
            OperationCanceledError: The operation was canceled by the user
        """
        self._validate_channel_list_or_raise(channels)

        suggested_parameters = []
        self._canceled = False
        self._last_event_channels = []
        self._bad_channels = []
        _update_progress(self.progress, 0, "Starting optimizer")
        try:
            self._backup_settings()
            self._set_default_dac_values()
            suggested_parameters = self._get_calibration_isel_for_channels(channels)
        except (OperationCanceledError, KeyboardInterrupt) as e:
            self.cancel()
            raise OperationCanceledError("Conversion Ramp tuning canceled") from e
        except Exception as e:
            self.cancel()
            raise OptimizationError("Optimization failed") from e
        finally:
            self._restore_settings()
            self._restore_tuning()

        return suggested_parameters

    def cancel(self):
        self._canceled = True

    def _reset_write_mask(self):
        self.channel_writer.reset_writemask()

    def _restore_tuning(self):
        """Restores isel ramp and cap to board.tuning values.

        If board.tuning doesn't contain tuning values, ramp
        and cap will be set to the default settings
        """
        conversion_ramp = self.board.tuning.get("conversion_ramp", None)
        if conversion_ramp:
            self.validate_iselrampcap_values(conversion_ramp)
            self._update_isel_for_channels(conversion_ramp)
        else:
            self._reset_tuning()

    def _reset_tuning(self):
        """Resets isel ramp and cap tuning parameters to defaults,
        ramp: 0x03, cap: 0x00.
        """
        self.channel_writer.write_analog_register("isel_cap_select", 0x00)
        self.channel_writer.write_analog_register("isel_ramp_current", 0x03)
        self.board.tuning["conversion_ramp"] = None
        self._reset_write_mask()

    def _backup_settings(self):
        self._backup_dac = self.board.dac_values.copy()

    def _restore_settings(self):
        for chan, value in self._backup_dac.items():
            get_dac_controller(self.board).set_single_dac(chan, value)

    def _set_default_dac_values(self):
        get_dac_controller(self.board).set_default_values()

    def _get_data(
        self, channels: list[int], num_events: int = 1, num_attempts: int = 5
    ):
        data_collector = get_data_collector(self.board)
        filter = partial(self._filter_data_for_channels, channels=channels)
        try:
            data = (
                data_collector.iter(num_events, num_attempts)
                .filter(filter, exclusion_limit=5)
                .collect()
            )
        except IterationError:
            raise BadDataError(self._iter_err_str(channels))

        return data

    def _filter_data_for_channels(self, evt, channels: list) -> bool:
        """Filter function to determine if an event has data for all the given channels"""
        self._last_event_channels = [c for c in channels if len(evt["data"][c]) != 0]
        no_data_channels = [c for c in channels if len(evt["data"][c]) == 0]
        if no_data_channels:
            LOGGER.warning("Could not capture data for channels: %s", no_data_channels)
        return True

    def _change_isel_reg_of_channel(self, channels: list, reg: str, val: int):
        """Wrapper function for writing the proper ISEL configuration to just the channels specified.

        Args:
            channels (list): Channels, relative to the chip, to write ISEL values to
            reg (str): Which ISEL parameter to change ("isel_cap_select" or "isel_ramp_current")
            val (int): Value to write (up to 3 for ramp, up to 15 for cap)
        """
        self.channel_writer.write_analog_register(reg, val, channels)

    def _get_calibration_isel_for_channels(self, channels: list[int]) -> dict:
        """
        Writes the proper ISEL ramp current and cap settings that result in an average event baseline
        value of target_value (default=500), for the chip specified.

        Args:
            chip (int): chip number to calibrate ISEL for

        Returns:
            suggested_vals (dict): the suggested parameters for isel ramp current and cap select.
            The key corresponds to the register to write, and value is a list of values per channel.
        """
        suggested_vals = {}

        suggested_vals_chip = self._get_suggested_value_for_channels(channels)

        for chan, val in zip(channels, suggested_vals_chip):
            if np.isnan(val).any():
                self._bad_channels.append(chan)
                continue

            suggested_vals[chan] = {
                "isel_cap_select": val[1],
                "isel_ramp_current": val[0],
            }
        if self._bad_channels:
            LOGGER.warning(self._chan_data_err_str(channels))

        return suggested_vals

    def _update_isel_for_channels(self, suggested_vals: dict):
        self.validate_iselrampcap_values(suggested_vals)
        for chan, suggestions in suggested_vals.items():
            for register, value in suggestions.items():
                self._change_isel_reg_of_channel([chan], register, value)
        self._reset_write_mask()

    def _get_suggested_value_for_channels(self, channels: list = range(16)) -> list:
        """
        Finds the linear parameters for each of the 3 ISEL ramp speeds by taking two points
        (all caps off and all caps on), finds the ramp speed that can result in an event average of 500,
        then finds the cap configuration within that ramp speed to ensure an event average of 500.

        Args:
            self.board (self.board): Upac96 self.board object with an active connection
            channels (list, optional): Chip-specific channels to configure. Defaults to range(16).

        Returns:
            list: the values written, 1 tuple (x) for each channel; x[0] is current, x[1] is ramp cap
        """

        channel_recs = []

        # find linear parameters for each of the 3 ISEL ramp speeds

        values = []

        isel_currents = [0, 1, 3]
        for step_idx, isel_current in enumerate(isel_currents):
            # Status message ###############################
            min_progress = 10
            max_progress = 80
            _update_progress(
                self.progress,
                10 + (max_progress - min_progress) * step_idx / len(isel_currents),
                f"Capturing Calibration Data:",
            )
            LOGGER.debug(
                "Capturing Calibration Data for isel_current_ramp: %d", isel_current
            )
            # Get slope between lowest and highest cap select values
            slope, yint = self._get_slope_for_current_state(isel_current, channels)
            # finds suggested cap select that result in an event average of 500
            value = self._get_suggested_cap_from_isel_lin(slope, yint)
            values.append(value)

        for channel in channels:
            if np.isnan(values[1][channel]).any():
                channel_recs.append(np.nan)
                continue

            # choose the midrange current as well as the cap for midrange current
            value_to_append = [1, values[1][channel]]

            # if cap configuration is too low, choose the lower current and its cap config.
            if values[1][channel] < 0:
                value_to_append = [0, values[0][channel]]

            # if cap configuration is too high, choose the higher current and its cap config.
            if values[1][channel] > 15:
                value_to_append = [3, values[2][channel]]

            # edge case handling in case cap config is still out of bounds
            value_to_append[1] = min(value_to_append[1], 15)
            value_to_append[1] = max(value_to_append[1], 0)

            channel_recs.append(value_to_append)

        return channel_recs

    def _get_suggested_cap_from_isel_lin(
        self, slopes: list[float], yints: list[float]
    ) -> list:
        """Suggests a ISEL cap configuration to write to given a the slope and y-intercept of the
        ADC-count and isel-cap linear relationship. Suggests one for each channel given. If the
        slope for a given channel is 0, the suggested isel-cap will be board default (0)

        Args:
            slopes (list(float)): List of calculated slope for ADC-count/isel-cap for each channel
            yints (list(float)): List of calculated y-int for ADC-count/isel-cap for each channel

        Returns:
            list: List of suggested cap configs to write to "isel_cap_select" digreg. Per channel
        """
        values = []
        for slope, yint in zip(slopes, yints):
            if np.isnan(slope) or np.isnan(yint):
                values.append(np.nan)
            elif slope != 0:
                values.append(round((self.target_value - yint) / slope))
            else:
                values.append(0)

        return values

    def _get_slope_for_current_state(
        self, current_state: int, channels: list = range(16)
    ):
        """Gets the slope and y-intercept from the ADC-mv/isel-cap relationship by taking events at the lowest and highest cap
        config. for a given ISEL current state.

        Args:
            current_state (int): ISEL current state to try (0, 1, or 3)
            channels (list): List of channels per chip to get linear parameters for

        Returns:
            list(float), float: list of slopes per channel and the y-intercept for the given current state
        """

        self._change_isel_reg_of_channel(channels, "isel_ramp_current", current_state)
        self._change_isel_reg_of_channel(channels, "isel_cap_select", 0)

        min_data = self._get_data(channels)
        chan_min = self._get_channel_avgs_from_readout(min_data)

        self._change_isel_reg_of_channel(channels, "isel_cap_select", 15)

        max_data = self._get_data(channels)
        chan_max = self._get_channel_avgs_from_readout(max_data)

        chan_slopes = [
            x / self._channels_per_chip for x in np.subtract(chan_max, chan_min)
        ]

        return chan_slopes, chan_min

    def _get_channel_avgs_from_readout(self, events: deque):
        """Gets the average ADC count per channel of a given event
        Will average if there are multiple events
        Args:
            readout (deque): Deque of parsed events to get ADC count averages from

        Returns:
            list(float): List of ADC count averages, per channel
        """
        chan_avgs = []

        events_array = np.array([event["data"] for event in events])
        averaged_events = np.mean(events_array, axis=0)

        for channel in averaged_events:
            chan_avgs.append(np.mean(channel))

        return chan_avgs

    def _validate_channel_list_or_raise(self, channels):
        """Validates a list of channels, and raises an error
        if there's a problem with them.

        Args:
            channels: the object that is supposedly a channels list

        Raises:
            TypeError if the channel list is not a list or
                values are not an int
            ValueError if the values are out of range
        """
        if not isinstance(channels, (list, range)):
            raise TypeError(
                f"Channels needs to be a list, not a {type(channels).__name__}"
            )
        if len(channels) > self._num_channels:
            raise ValueError(f"Too many channels, max is {self._num_channels}")
        for channel in channels:
            if not isinstance(channel, int):
                raise TypeError(f"Channels can only contain integers")
            if channel >= self._num_channels or channel < 0:
                raise ValueError(f"Channel {channel} is out of bounds")

    def validate_iselrampcap_values(self, tuning_values: dict):
        """Validates parameters for isel ramp and cap. Ensures the channels,
        and values for ramp and cap are valids.

        Args:
            tuning_values (dict): the tuning parameters for isel ramp & cap

        Raises:
            TypeError if the parameters are not of type dict
            ValueError if the channel or tuning parameters are out of range
            KeyError if the parameters holds an invalid key for isel ramp & cap
        """
        cap_width = self.board.registers["analog_registers"]["isel_cap_select"][
            "bitwidth"
        ]
        ramp_width = self.board.registers["analog_registers"]["isel_ramp_current"][
            "bitwidth"
        ]
        max_values = {}
        max_values["isel_cap_select"] = 2**cap_width - 1
        max_values["isel_ramp_current"] = 2**ramp_width - 1
        if not isinstance(tuning_values, dict):
            raise InvalidOptimizationParameters(
                f"Conversion Ramp parameters should be of type dict, got type: {type(tuning_values).__name__}"
            )

        if len(tuning_values) > self._num_channels:
            raise InvalidOptimizationParameters(
                f"Dictionary contains more than {self._num_channels} channel values"
            )

        for channel, suggested_tuning in tuning_values.items():
            if channel < 0 or channel > self._num_channels:
                raise InvalidOptimizationParameters(
                    f"Channel {channel} is out of range"
                )

            for key, value in suggested_tuning.items():
                if not isinstance(value, int):
                    raise InvalidOptimizationParameters(
                        f"Tuning values should be of type int, got type: {type(value).__name__}"
                    )
                if key not in ["isel_ramp_current", "isel_cap_select"]:
                    raise InvalidOptimizationParameters(f"{key} Not in tuning file")
                if value < 0 or value > max_values[key]:
                    raise InvalidOptimizationParameters(
                        f"Value of {value} for key {key} is out of range"
                    )

    def _iter_err_str(self, channels: list[int]) -> str:
        """Get a string suitible for an error message when an iteration error occurs.

        Includes any missing channels and chips.
        """
        channels_per_chip = self.board.params.get("channels_per_chip", 16)
        missing_channels = find_missing_channels(channels, self._last_event_channels)
        missing_channels = group_channels_by_chip(missing_channels, channels_per_chip)

        chip_parts = []
        for chip, missing_in_chip in missing_channels.items():
            if len(missing_in_chip) == 0:
                continue
            msg = f"- Chip {chip}: all channels"
            if len(missing_in_chip) < channels_per_chip:
                chan_repr = ", ".join(map(str, missing_in_chip))
                msg = f"- Chip {chip}: channel(s) {chan_repr}"
            chip_parts.append(msg)

        missing_formatted = "\n".join(chip_parts)
        return (
            "Failed to generate ALO tuning because the following chips/channels "
            "cannot be read from the hardware:\n"
            f"{missing_formatted}\n\n"
            "Please try again and power cycle the board if the issue persists. To instead ignore this error, "
            "deselect the affected channel(s) and try again."
        )

    def _chan_data_err_str(self, channels: list[int]) -> str:
        """Get a string suitible for an error message when unable to capture
        data for some channels. States which channels have been tuned, and
        which ones have failed.
        """
        channels_per_chip = self.board.params.get("channels_per_chip", 16)
        missing_channels = find_missing_channels(channels, self._last_event_channels)
        missing_channels = group_channels_by_chip(missing_channels, channels_per_chip)
        tuned_channels = group_channels_by_chip(channels, channels_per_chip)
        tuned_chip_parts = []
        missing_chip_parts = []
        for chip, missing_in_chip in missing_channels.items():
            if len(missing_in_chip) == 0:
                continue
            for missing_chan in missing_in_chip:
                tuned_channels[chip].remove(missing_chan)
            msg = f"- Chip {chip}: all channels"
            if len(missing_in_chip) < channels_per_chip:
                chan_repr = ", ".join(map(str, missing_in_chip))
                msg = f"- Chip {chip}: channel(s) {chan_repr}"
            missing_chip_parts.append(msg)

        for chip, tuned_in_chip in tuned_channels.items():
            if len(tuned_in_chip) == 0:
                continue

            msg = f"- Chip {chip}: all channels"
            if len(tuned_in_chip) < channels_per_chip:
                chan_repr = ", ".join(map(str, tuned_in_chip))
                msg = f"- Chip {chip}: channel(s) {chan_repr}"
            tuned_chip_parts.append(msg)

        missing_formatted = "\n".join(missing_chip_parts)
        tuned_formatted = "\n".join(tuned_chip_parts)
        return (
            "Tuned chips/channels:\n"
            f"{tuned_formatted}\n\n"
            "Failed to generate ALO tuning for the following chips/channels\n"
            f"{missing_formatted}\n\n"
            "Please try again and power cycle the board if the issue persists."
            "deselect the affected channel(s) and try again."
        )


def _update_progress(receiver, percent: float, message: str):
    """Updates a progress "receiver" with a percent and message.

    The receiver can be a list or deque of (percent, message) tuples,
    or a ProgressDialog (for use with NaluScope).

    Args:
        receiver (list, deque, or ProgressDialog): the receiver of the progress update.
            Can be None to report nothing
        percent (float): the percent completion of the task
        message (str): a description of what is currently taking place
    """
    if receiver is None:
        LOGGER.debug("%s | %s", percent, message)
        return
    elif isinstance(receiver, (list, deque)):
        receiver.append((percent, message))
    else:
        try:
            receiver.update_status(percent, message)
        except:
            pass



def get_usb_reader(board, serial_buffer, *args, **kwargs):
    """Get a usb reader depending on connection type"""
    conn_type = board.connection_info.get("type", "uart")

    if conn_type in ["ft60x"]:
        from naludaq.daq.workers.worker_usb_reader import UsbReader

        worker = UsbReader(
            board.connection,
            serial_buffer,
            *args,
            frequency=kwargs.get("frequency", 1000),
            **kwargs,
        )
    else:
        from naludaq.daq.workers.worker_serial_reader import SerialReader

        worker = SerialReader(
            board.connection,
            serial_buffer,
            *args,
            **kwargs,
        )
    return worker



import logging
import time
from collections import defaultdict
from typing import Dict, Iterable

import numpy as np

from naludaq.controllers.board import get_board_controller
from naludaq.helpers import type_name

LOGGER = logging.getLogger("naludaq.threshold_scan")
NUM_NO_TRIGGERS = 5


class ThresholdScan:
    """Tool for sweeping over trigger thresholds and counting number of triggers.
    Can be used to determine the ideal point to set the trigger.
    """

    def __init__(
        self,
        board,
        channels: Iterable[int] = None,
        scan_values: Iterable[int] = None,
    ):
        """Class to run ThresholdScan. Used to find the number of
        triggers for each trigger threshold.

        Args:
            board (Board): Board to run a threshold scan on
            threshold_values (Iterable[int]): threshold values to run the scan for.
        """
        self.board = board
        self.channels = channels
        self._cancel = False
        self._progress: list = []
        self._num_trig_chans = board.channels
        self.scan_values = scan_values
        self._counters = defaultdict(int)

        self._board_controller = get_board_controller(self.board)

    @property
    def channels(self) -> list[int]:
        """Get/set the channels to run the scan on."""
        return self._channels.copy()

    @channels.setter
    def channels(self, channels: Iterable[int]):
        if channels is None:
            channels = range(self.board.channels)
        if not isinstance(channels, Iterable):
            raise TypeError(
                f"Channels must be a Iterable[int], not {type_name(channels)}"
            )
        if any(not isinstance(c, int) for c in channels):
            raise TypeError("Channels must be a Iterable[int]")
        if any(not 0 <= c < self.board.channels for c in channels):
            raise ValueError("One or more channels is out of bounds")
        self._channels = list(channels)

    @property
    def scan_values(self) -> np.ndarray:
        """Gets/sets the threshold values to read the scalars for."""
        return self._threshold_values.copy()

    @scan_values.setter
    def scan_values(self, values: Iterable[int]):
        if values is None:
            values = self._default_scan_values()
        if not isinstance(values, Iterable):
            raise TypeError(f"Values must be a Iterable[int], not {type_name(values)}")
        values = np.array(values)
        if (
            np.min(values) < self.board.trigger._min_thresholds
            or np.max(values) > self.board.trigger._max_thresholds
        ):
            raise ValueError("One or more values is out of bounds")
        self._threshold_values = values

    def run(self, pause: float = 0.1):
        """Scan the scalars at the trigger values defined by the scan_values property.

        Returns the amount of hits for each scanned trigger value in the range.

        Args:
            pause (float): amount of seconds to pause in between samples.

        Returns:
            Tuple of (trigger value per channel as `np.array`, `irange` as list).
        """
        self._cancel = False
        self._progress.append((0, "Backing up board settings"))
        self._backup_board_settings()
        output = self._get_scalar_values(pause)
        self._progress.append((95, "Restoring board settings"))
        self._restore_board_settings()
        self._progress.append((100, "Done"))

        return (np.array(output), self.scan_values.copy())

    def _backup_board_settings(self):
        """Saves a copy of board trigger values to be restored later."""
        self._original_trigger_values = self.board.trigger.values.copy()
        self.chan_bk = self.channels.copy()

    def _restore_board_settings(self):
        """Restores the original trigger values from board."""
        self.board.trigger.values = self._original_trigger_values
        self.board.trigger.write_triggers()
        self.channels = self.chan_bk

    def _get_scalar_values(self, pause: float) -> np.ndarray:
        """Scan the range and return np.array with trigger amounts

        Args:
            pause (float): amount of seconds to pause in between samples.

        Returns:
            Triggers value per channel as `np.array`.
        """
        scan_values = self.scan_values
        self._counters = defaultdict(int)

        output = np.zeros((self.board.channels, len(scan_values)))
        for i, value in enumerate(scan_values):
            if self._cancel:
                break
            LOGGER.debug("Scan progress: %s/%s", i + 1, len(scan_values))
            self._progress.append(
                (int(95 * i / len(scan_values)), f"Scanning {(i+1)}/{len(scan_values)}")
            )
            time.sleep(pause)

            scals = self._get_scaler_value(value)

            for chan in self.channels:
                output[chan][i] = scals[chan]

            self.check_early_stopping(scals)
        return output

    def cancel(self):
        """Cancels the threshold scan if one is currently running.

        This function must be called from a different thread, as the
        threshold scan is blocking. The threshold scan will stop as
        soon as possible.
        """
        self._cancel = True

    def _set_trigger_thresholds(self, channel: int, trigger_value: int):
        trigger_vals = {ch: 0 for ch in range(self.board.channels)}
        trigger_vals[channel] = int(trigger_value)
        self.board.trigger.values = trigger_vals

    def _get_scaler_value(self, trigger_value: list[int]) -> Dict[int, int]:
        """Read the scalers for selected channels for a specific trigger value.

        Args:
            channels (list): list of channels to scan the scalers

        Returns:
            dict of scaler values for the selected channels
        """
        scalers = {}
        for chan in self.channels:
            self._set_trigger_thresholds(chan, trigger_value)
            scal = self._board_controller.read_scalar(chan)
            scalers[chan] = scal

        return scalers

    @property
    def progress(self):
        """Get/Set the progress message queue.

        This is a hook to read the progress if running threads.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    def _default_scan_values(self) -> np.ndarray:
        """Get the default scan values from the params"""
        scan_params = self.board.params.get("threshold_scan", {})
        return np.arange(
            scan_params.get("start", 500),
            scan_params.get("stop", 3500),
            scan_params.get("stepsize", 5),
        )

    def check_early_stopping(self, scals: dict[int, int]):
        """Check if the scan should be stopped early.

        Stops the scan to a channel if there has been a a trigger.
        If there are no channels left to scan a cancel is set.

        Args:
            scals (Dict[int, int]): Scaler values for each channel.
        """
        for i, val in scals.items():
            if val > 0:
                self._counters[i] = NUM_NO_TRIGGERS
            if self._counters[i] > 0:
                self._counters[i] -= 1
                if self._counters[i] == 0:
                    channels = self.channels
                    channels.remove(i)
                    self.channels = channels
                    if len(self.channels) == 0:
                        self._cancel = True



"""

"""
from abc import ABC, abstractmethod
from typing import List

from naludaq.communication import (
    AnalogRegisters,
    ControlRegisters,
    DigitalRegisters,
    I2CRegisters,
)
from naludaq.helpers.exceptions import InvalidBoardModelError


class Initializers(ABC):
    def __init__(self, board, expected_models: List[str]) -> None:
        """Base class for board initializers.

        Args:
            board (Board): the board to initialize.
            expected_models (List[str]): a list of valid models this
                initializer will work for. This should be determined
                by the subclass.

        Raises:
            InvalidBoardModelError if the board is of an incorrect type.
        """
        if isinstance(expected_models, str):
            expected_models = [expected_models]
        if board.model not in expected_models:
            raise InvalidBoardModelError(
                f'Incorrect initializer used for board "{board.model}"'
            )
        self.board = board
        self.analog_registers = AnalogRegisters(self.board)
        self.control_registers = ControlRegisters(self.board)
        self.digital_registers = DigitalRegisters(self.board)
        self.i2c_registers = I2CRegisters(self.board)

        self.control_write = self.control_registers.write
        self.digital_write = self.digital_registers.write
        self.analog_write = self.analog_registers.write
        self.i2c_write = self.i2c_registers.write

    @abstractmethod
    def run(self) -> bool:
        """Runs the initialization sequence for the board.

        Returns:
            True if successful, False otherwise.
        """
        pass



"""Parser for HiPER multichip data
"""

import logging

import numpy as np

from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers.headers import get_header_parser

LOGGER = logging.getLogger("naludaq.parsers.hiper")


# check for chips in data. GEt the location on the first of each of the chips
# Multi-cip functions
class HiperParser:
    """Class to repackage HiPER multichip data.

    Use instead of the normal packager
    """

    def __init__(self, params):
        self.params = params

        # TODO(Marcus): Replace with aardvarc header parser
        self._parse_event_headers = get_header_parser(self.params["model"])

        # TODO(Marcus): connect all the params to the input.
        self.test_amount_windows = 3
        self.samples = self.params.get("samples", 64)
        self.windows = self.params.get("windows", 512)
        self.channels = self.params.get("channels", 4)
        self.num_evt_headers = self.params.get("num_evt_headers", 3)
        self.num_channel_headers = self.params.get("num_channel_headers", 1)
        self.num_window_headers = self.params.get("num_window_headers", 1)
        self.data_bitmask = self.params.get("data_bitmask", 8191)
        self.chanmask = self.params.get("chanmask", 3072)
        self.windmask = self.params.get("windmask", 1022)
        self.chanshift = self.params.get("chanshift", 10)
        # self.num_footer_words = self.params.get("num_footer_words", 2)
        # self.last_bits = self.params.get("lastbits", 1)
        self.chan_step_size = self.samples + self.num_channel_headers
        self.chips = 14
        self.channels_per_chip = 4
        self.channels = self.channels_per_chip * self.chips
        self.samples = 64
        self.splitwords = {
            bytes.fromhex(f"bbb{hex(x)[2:]}"): bytes.fromhex(f"fff{hex(x)[2:]}")
            for x in range(self.chips)
        }

        self._valid_splits = []

    @property
    def valid_splits(self):
        return self._valid_splits

    @valid_splits.setter
    def valid_splits(self, splits):
        self._valid_splits = []
        for spl in splits:
            self._valid_splits.append(bytes.fromhex(f"bbb{hex(spl)[2:]}"))

    def _validate_input_package(self, in_data):
        """Validate the input data.
        Raises:
            BadDataException if:
            - there is no data.
            - there is not a en event amount of bytes, since 8-bits are combined to 16-bit words.
            - the data is not divisable by sample_num + window_headers.
        """
        # if not isinstance(in_data, dict):
        #     raise TypeError("raw_data is not a dict it's a %s", type(in_data))

        # if in_data['rawdata'] == b"":
        #     raise BadDataError('No Data Found')

        # # Edge case, data doesn't contain valid words.
        # if len(in_data['rawdata']) % 2 != 0:
        #     raise BadDataError(
        #         f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
        #         f"length is not an even amount of words, "
        #         f"len {len(in_data['rawdata'])}"
        #     )

    def parse(self, indata):
        """Takes a set of binary data and split&parse it to events"""
        # data["rawdata"][0:24].hex()
        if indata.get("rawdata", None):
            data = indata["rawdata"]
        else:
            raise BadDataError("Datapackage is invalid")
        output = []

        if not self.valid_splits:
            valid_splits = self._check_for_chips(data)
        else:
            valid_splits = self.valid_splits
        # print(f"valid splits: {valid_splits}")
        if valid_splits:
            split_data = self.split_data_into_packs(data, valid_splits[0])
        else:
            split_data = [indata["rawdata"]]
        # print(f"Len of data: {len(split_data)}")
        # put in foor loop
        all_chippacks = []
        for idx, split in enumerate(split_data):
            # print(f"Len of split: {len(split)}")
            all_chippacks.append(self.split_packs_into_chippacks(split, valid_splits))

        # this needs to be done in a loop too:

        for chippack in all_chippacks:
            # print(f"Len of chippack: {len(chippack)}")
            eventpack = self.split_chippacks_into_events(chippack)
            self.parse_eventpacks(eventpack, output)
        # print(f"Parsed package: {output[0]}")
        try:
            output[0]["event_num"] = indata["pkg_num"]
            return output[0]
        except Exception:
            indata["event_num"] = indata["pkg_num"]
            return indata

    def _check_for_chips(self, idata) -> list:
        """Checks which chips are present in the data.
        Args:
        Returns:
            the (start,end) string tuple for the available chips.
        """
        p = -1
        valid_splits = []
        for spl, _ in self.splitwords.items():
            p = idata.find(spl)
            if p != -1:
                valid_splits.append(spl)
        return valid_splits

    def _extract_chipnum(self, data: bytearray) -> str:
        """Extract the hex chipnum from data"""
        return data[0:2].hex()[-1]

    # Find all splits for the first chip available, creating the packs with events
    # each following chip will contain the following pieces of the events.
    def split_data_into_packs(self, idata, splitword):
        """Takes all the indata and splits into packs
        A pack is from the header of the first(lowest number) chip until it's found again.
        Args:
            idata: All captured data.
            splitword: header of the lowest numbered chip
        Returns:
            list of packs to process.
        """
        splits = []
        running = True
        p = idata.find(splitword)
        while running:
            p2 = idata.find(splitword, p + 1)
            if p2 == -1:
                p2 = len(idata)
                running = False
            splits.append(idata[p:p2])
            p = p2
        return splits

    # for each pack, extract the data for each chip
    # that extracted data will later be reassembled into events.
    # Split the pack
    def split_packs_into_chippacks(self, datpack, splitwords):
        chippacks = []
        for splitstart in splitwords:
            splitend = self.splitwords[splitstart]
            p = datpack.find(splitstart)
            p2 = datpack.find(splitend)
            if p != -1 and p2 != -1:
                chippacks.append(datpack[p:p2])

        return chippacks

    # split chippack into events
    def split_chippacks_into_events(self, chippacks):
        """Split the chippacks into individual events
        The events are then stored in a dict with their chip number for reassebly.
        Args:
            all_chippacks (list[list]): All chippacks in a list, each list in the list contains data from all each chip.
        Returns:
            {chipnum(hex): [events]}
        """
        event_footer = bytes.fromhex("a5af5afc3c3c")
        eventpack = {
            hex(x)[2:]: [] for x in range(self.chips)
        }  # TODO: Not needed parse and reassmeble in one step

        for idx, chip in enumerate(chippacks):
            # extract chip number
            try:
                chipnum = self._extract_chipnum(chip)
            except Exception:
                LOGGER.error(
                    f"Failed, failing idx: {idx}/{len(chippacks)}, data: {chip}"
                )
            # split the events
            events = [
                x[2:] for x in chip.split(event_footer) if len(x) > 0
            ]  # OR always toss last piece? either incomplete data or no data? IMPORTANT! FIGURE OUT!

            eventpack[chipnum] = events

        eventpack = {k: v for k, v in eventpack.items() if len(v) > 0}
        return eventpack

    # parse the data
    def parse_eventpacks(self, eventpack, output):
        """Parses eventpacks into deque of events"""
        idx = 0

        while True:
            event = {}
            for k in eventpack.keys():
                try:
                    next_pack = eventpack[k][idx]
                except Exception:
                    continue
                else:
                    try:
                        event = self.parse_digital_data(next_pack, k, event)
                    except Exception as e:
                        LOGGER.error(f"Parsing data failed: {e}")
                        continue
                    event["time"] = self._add_xaxis_to_event(event)

            if not event:
                break
            output.append(event)
            idx += 1

    def parse_digital_data(self, in_data, chipid, event={}):
        """Print the header for each end every word in the data.
        Warning, long list...
        Args:
            data (np.array or list): all the data, not multi dimensional so one channel at the time.
        """
        # TODO(Marcus): Need logic for the different channels. Since this needs to be processed from readout data.
        self.samples
        chipchannels = self.channels_per_chip
        self.chips
        chipid = int(chipid, 16)
        channels = self.channels

        hex_data = in_data.hex()
        header_len = 9
        stepsize = 65 * 3
        split_word = "FC3C3C"
        readout_data = hex_data[header_len:]
        if not event:
            event = {
                "header": hex_data[:header_len],
                "data": [[] for _ in range(channels)],
                "window_labels": [[] for _ in range(channels)],
                "start_window": int(readout_data[0:3], 16) & (2**7 - 1),
            }

        for block in readout_data.upper().split(split_word):
            # chip_header = block[:4]

            for idx in range(0, len(block) - stepsize, stepsize):
                header = block[idx : idx + 3]
                chan = (int(header, 16) >> 9) & 3
                channel = (chipid * chipchannels) + chan
                window = int(header, 16) & (2**7 - 1)
                pack = block[idx + 3 : idx + stepsize]
                windata = [self._hex_comb(x, pack) for x in range(0, len(pack), 3)]

                event["data"][channel].extend(windata)
                event["window_labels"][channel].append(window)
        event["start_window"]
        # event['data'] = np.array(event['data'])
        event["event_num"] = 0

        return event

    @staticmethod
    def _hex_comb(index, data):
        """Combines hexnibbles into 12bit data"""
        outp = (
            (int(data[index], 16) << 8)
            | (int(data[index + 1], 16) << 4)
            | (int(data[index + 2], 16))
        )
        return outp

    def _add_xaxis_to_event(self, event):
        """Adds an x-axis to the data.
        Based on the amount of channels and samples it will add a timeaxis
        per channel to the event dict.
        Returns:
            numpy array with sample numbers for each channel.
        """
        times = list()
        samples = self.samples
        channels = self.channels
        start_window = event["start_window"]
        for chan in range(0, channels):
            # LOGGER.debug("winds=%s", event["window_labels"][chan])
            winds = np.array(event["window_labels"][chan])
            winds += self.windows * (winds < start_window)
            winds -= start_window

            tim = (samples * np.repeat(winds, samples)) + (
                np.ones((len(winds), samples), dtype=int)
                * np.arange(0, samples, dtype=int)
            ).flatten()

            times.append(tim)

        return times



"""Analog registers
===================
Registers controlling the analog part of the ASIC.

What are the analog registers?
It's a set of registers on the FPGA matching a set of registers on the chip
By changing these registers it controls the analog portion of the chip.

Command structure:
------------------
Updating the analog command uses a format 'BXMAAVVV'

- B tells the FPGA command parser to expect a analog/digital command.
- M is mode, 0 = write, 1 = read
- AA = address
- VVV = value

Will be interpereted by the FPGA as hex. it's 32-bit = 8 hex values.

Functions:
----------
- write_all
- write
- write2addr


"""
from logging import getLogger

from naludaq.communication._chip import ChipRegisters
from naludaq.communication.digital_registers import DigitalRegisters

LOGGER = getLogger(__name__)


class AnalogRegisters(ChipRegisters):
    def __init__(self, board, chips: "int | list[int]" = None):
        """Registers controlling the analog part of the ASIC(s).

        For a more thorough explanation of the multichip aspect and its implications,
        see the `multichip` module.

        Args:
            board (Board): board object
            chips (int | list[int] | None): chip(s) to write to. If None, all chips are assumed.
        """
        super().__init__(board, "analog_registers", chips)
        self._write_cmd = 0xB00
        self._width_cmd = 3
        self._width_addr = 3  # hex char
        self._width_val = 3  # hex char

        self._fix_timing = self._select_fixtiming_method()

    def _select_fixtiming_method(self):
        """Select the timing fix method depending on the board model.

        Some boards have a different wiring and require a different caluclation.

        Returns:
            timing correction function for the current board.
        """
        if self.board.model in ["hdsocv2_eval", "hdsocv2_evalr2"]:
            return self._fix_timing_even
        return self._fix_timing_odd

    # READ #########################################################

    def _addr_value(self, addr: str, chip: int):
        """Reimplemented to correct strobe values."""
        if self.board.params.get("strobe_values_correction", False):
            result = self._correct_strobes(addr, chip)
        else:
            result = super()._addr_value(addr, chip)
        return result

    def _correct_strobes(self, addr: str, chip: int) -> int:
        """Generate the value from the hex address.

        An address can contain multiple registers.
        Generate the value depending on the bit positions of all registers on the address.

        Args:
            addr (str): Hex value of the address, 2 hex char long.

        Returns:
            int: value of the combined registers.

        Raises:
            AttributeError if the address is not valid.
        """
        addr = addr.upper()
        strobe_correction_keys = self.board.params.get("strobe_correction_keys", [])
        value = 0
        for name, register in self.registers_on_address(addr).items():
            if not isinstance(register["value"], list):
                register["value"] = [register["value"]]
            reg_value = register["value"][chip]
            reg_pos = register["bitposition"]
            reg_width = register["bitwidth"]

            # FIX TIMING REGS, ACCORDING TO BOARD PARAMETER timing_correction
            if name in strobe_correction_keys:
                reg_value = self._fix_timing(reg_value)
            value += 2**reg_pos * (int(reg_value) % (2**reg_width))
        return value

    def _generate_write_atom(self, addr: str, value: int) -> str:
        """Generate the string to send to the board.

        OVERLOAD the standard way of generating the command since the analog command parsing
        sometimes

        Abstracts the model differences away from the main code, AARDVARCv2 got a timing hack.

        Args:
            board(obj)
            name(str): name of the analog register
            value(): Value to set the register to.
        """
        write_addr = f"{self._write_cmd:X}"
        cmd_addr = addr[1:] if len(addr) == 3 else addr
        set_ext = restore_ext = ""
        if len(addr) == 3 and addr[0] == "1":
            set_ext, restore_ext = self._generate_ext_bit_commands()
        if isinstance(value, (int, bool)):
            value = f"{value:X}".zfill(self._width_val)
        command = f"{write_addr}{cmd_addr}{value}"  # First 5 hex chars, address should be 3 hex. B001c
        return f"{set_ext}{command}{restore_ext}"

    def _fix_timing_odd(self, in_value: int, *args, **kwargs):
        """The timing signal data mux is all mixed up for simplicity of routing, so this fixes it

        Don't use unless you know what it does. Contains hardcoded values.
        """
        samples = self.board.params["samples"]

        if in_value < samples:
            fixed = in_value * 2 + 1
        else:
            fixed = 2 * ((2 * samples) - in_value - 1)
        return fixed

    def _fix_timing_even(self, in_value: int, *args, **kwargs):
        """The timing signal data mux is all mixed up for simplicity of routing, so this fixes it

        Don't use unless you know what it does. Contains hardcoded values.
        """
        samples = self.board.params["samples"]

        if in_value < samples:
            fixed = in_value * 2
        else:
            fixed = 2 * ((2 * samples) - in_value - 1) + 1
        return fixed

    def _generate_ext_bit_commands(self) -> tuple[str, str]:
        """Generate commands to set or clear the external bit.

        Args:
            board (Board): the board
            wait (bool): whether to include a wait command
            restore (bool): whether to include a restoration command

        Returns:
            tuple[str, str]: set, clear commands
        """
        dr = DigitalRegisters(self.board)
        set_cmd = dr.generate_write("address_ext", True)
        clear_cmd = dr.generate_write("address_ext", False)
        return set_cmd, clear_cmd



from logging import getLogger
from typing import MutableSequence

from naludaq.daq.workers.packager.worker_packager import PackagerLight

MAX_TIMEOUT = 50

logger = getLogger("naludaq.hiper_apckager")


class HiperPackager(PackagerLight):
    def __init__(
        self,
        input_buffer: MutableSequence,
        output_buffer: MutableSequence,
        output_answers: MutableSequence,
        stop_word: bytes,
        frequency: int,
    ):
        super().__init__(
            input_buffer=input_buffer,
            output_buffer=output_buffer,
            output_answers=output_answers,
            stop_word=stop_word,
            frequency=frequency,
        )

        self.answer_size = 8
        self._max_chip = -1
        self._prev_split = -1  # reset counter
        self._prev_buflen = -1
        self._prev_increase = 0
        self._timeout = 0

        self.chips = 14
        self.splitwords = [
            (x, bytes.fromhex(f"bbb{hex(x)[2:]}"), bytes.fromhex(f"fff{hex(x)[2:]}"))
            for x in range(self.chips)
        ]

    def run(self):
        logger.debug("Trying to start packager.")
        self._package_worker()

    def stop(self):
        self.running = False

    def _route_serial_data(self, internal_buffer) -> bytearray:
        """Splits and routes serial data into the appropriate
        output buffers.

        Args:
            buff (bytearray): the serial input buffer

        Returns:
            Any leftover serial data following the last stop word.
        """
        valid_splits = self._check_for_chips(internal_buffer)
        if not valid_splits:
            return internal_buffer
        max_chip, stop_word = valid_splits[
            -1
        ]  # Biggest chipnum is last in list by design, if this increased then flag stays low
        trigger = False
        if max_chip == self._prev_split:  # no increase in chip ending
            if (
                len(internal_buffer) >= self._prev_buflen + self._prev_increase
                and self._timeout > MAX_TIMEOUT
            ):  # we knew it took _prev_buflen to find an increase previously  # Can cause issue if polling too fast and no new endwords are polled, add a timeout or data amount recv critera?
                trigger = True
                self._timeout = 0
            if (
                len(internal_buffer) == self._prev_buflen
                and self._timeout > MAX_TIMEOUT
            ):  # No more data received.
                trigger = True
                self._timeout = 0
            else:
                self._timeout += 1
        else:
            # Will capture first time the last chipend is found
            self._prev_increase = (
                len(internal_buffer) - self._prev_buflen
            )  # will give us the requirec increase to find a chipend.
            self._prev_buflen = len(internal_buffer)
            self._timeout += 1

        if trigger:
            packages, answers, rest = self._slice_buffer(internal_buffer, stop_word)
            for package in packages:
                self.output_buffer.append(package)
            for package in answers:
                self.answers.put(package)

            self._prev_split = -1  # reset counter
            self._prev_buflen = len(rest)
            trigger = False
        else:
            # Not done collecting data yet
            self._prev_split = max_chip
            self._prev_buflen = len(internal_buffer)
            rest = internal_buffer

        return rest

    def _check_for_chips(self, idata) -> list:
        """Checks which chips are present in the data.
        Args:
        Returns:
            the (start,end) string tuple for the available chips.
        """
        p = -1
        valid_splits = []
        for idx, spl, sp2 in self.splitwords:
            p = idata.find(sp2)
            if p != -1:
                valid_splits.append((idx, sp2))
        return valid_splits

    def _slice_buffer(self, internal_buffer, stopword):
        """Slice the internal buffer at the stopword.

        Takes an internal buffer as a bytearray and cuts it at the stopword.
        The output is a list of packages.
        Once sliced the last part will be the first part of the next batch.

        Args:
            internal_buffer (bytearray): raw data to slice.

        Returns:
            output: a list of packages
            rest: whatever is after the last stopword.
        """

        packages = internal_buffer.split(stopword)

        output = [
            self._create_data_package(x + stopword, self._pkg_num + idx)
            for idx, x in enumerate(packages[:-1])
            if self._is_data_package(x)
        ]
        answers = [
            self._create_answer_package(x + stopword)
            for x in packages[:-1]
            if self._is_answer_package(x)
        ]
        self._pkg_num += len(packages) - 1
        rest = packages[-1]
        del packages

        return output, answers, rest



"""Initializer for the UDC board

This initializer will run the init sequence for the UDC board.
It contains all UDC16 specific startup sequence
"""
import functools
import logging
import time

from naludaq.board import initializers
from naludaq.controllers.biasing_mode import get_biasing_mode_controller

logger = logging.getLogger(__name__)


def func_print(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logger.info(f" Running: {func.__name__} ".ljust(80, "#"))
        v = func(*args, **kwargs)
        return v

    return wrapper


class InitUDC16(initializers.Initializers):
    def __init__(self, board):
        super().__init__(board, "udc16")
        self.power_sequence = [
            "2v5_en",
            "1v2_en",
            "clk2v5_en",
            "clk1v8_en",
        ]
        self.rxtx = [
            "rx_enable",
            "tx_enable",
        ]

    def run(self):
        # UDC_FPGA_register_init
        self._fpga_init()

        # POWER ON FMC BOARD
        self._power_toggle(False)
        self._power_toggle(True)

        self._program_clock()

        # reset sysrst and regclr
        self._reset_sys()
        self._reset_digital()

        # Enable Rx and Tx
        self._rxtx_toggle(True)

        # Analog and digital startup:
        # Digital startup
        self._digital_startup()

        # Analog_startup
        self._analog_startup()

        # UDC ASIC reset
        self._asic_reset()

        # Set DACS
        self._init_dacs()

        return True

    @func_print
    def _fpga_init(self):
        """Write all FPGA registers to their default values."""
        self.control_registers.write_all()
        self.i2c_registers.write("i2c_en", True)

    @func_print
    def _power_toggle(self, state):
        """Toggles the power rails defined in `power_sequence` attribute"""
        for register in self.power_sequence:
            self.control_write(register, state)

    @func_print
    def _program_clock(self):
        """Program the clock chip, uses the clockfile in the paramters if it exits."""
        time.sleep(0.05)
        self.board.clock.program(filename=self.board.params["clock_file"])
        time.sleep(0.05)

    @func_print
    def _reset_sys(self):
        """Reset the ASIC and clear the analog and digital registers"""
        self.control_write("sysrst", True)
        self.control_write("regclr", True)

        self.control_write("sysrst", False)
        self.control_write("regclr", False)
        self._reset_digital()

    def _reset_digital(self):
        """Reset digital side of the ASIC"""
        self.control_write("digrst", True)
        self.control_write("wave_fifo_reset", True)

        self.control_write("digrst", False)
        self.control_write("wave_fifo_reset", False)

    @func_print
    def _asic_reset(self):
        """Reset the asic, analog side, digital side, and fifo"""
        self.control_write("sysrst", True)
        self.control_write("sysrst", False)
        self._reset_digital()

    @func_print
    def _rxtx_toggle(self, state):
        """Toggles ASIC rx and tx enable. True when talking to ASIC."""
        for register in self.rxtx:
            self.control_write(register, state)

    @func_print
    def _digital_startup(self):
        """Startup the digital side by programming all registers"""
        self.digital_registers.write_all()

    @func_print
    def _analog_startup(self):
        """Start the analog side of the chip"""
        # PCLK decoding ANALOG REGISTERS decoding
        self.analog_registers.write_all()

        # DLL START UP SEQUENCE
        self._dll_startup()
        self._set_default_biasing_mode()

    @func_print
    def _dll_startup(self):
        """Starting the delay line on the analog side."""
        self.analog_write("qbias", 0)
        self.analog_write("vadjn_sw", 1)
        time.sleep(1)
        self.analog_write("qbias", 2048)
        self.analog_write("vadjn_sw", 0)

    @func_print
    def _init_dacs(self):
        """Program the left and right side DACs"""
        dacvalues = self._get_dacvalues()

        for channel, value in dacvalues.items():
            self.board.ext_bias.set_single_dac(channel, value)

    def _get_dacvalues(self) -> dict:
        """Return a dict with all the "channels: dac_values"""
        return self.board.params.get("ext_dac", {}).get("channels", {})

    def _set_default_biasing_mode(self):
        get_biasing_mode_controller(self.board).set_biasing_mode(
            self.board.params["intamp_bias_mode_chan0_7"],
            self.board.params["intamp_bias_mode_chan8_15"],
        )



"""Readout settings for the hdsocv1 board.

Controls the readout channels and the readout windows.
"""
from logging import getLogger
from typing import List

from naludaq.communication import AnalogRegisters

from .default import ReadoutController

LOGGER = getLogger("naludaq.readout_controller_hdsoc")


class HDSoCReadoutController(ReadoutController):
    """Readout controller for the hdsocv1 board.

    Controls the readout channels and the windows, lookback, and write after trigger.
    """

    def __init__(self, board):
        """Readout Controller for HDSoCv1.

        Args:
            board (Board): the board object.
        """
        super().__init__(board)

    def set_readout_channels(self, channels_to_read: List[int]):
        """Select channels to readout.

        Will turn on the chips necessary to read out from the given channels
        (e.g. channels [15, 16, 17] -> chips [0, 1]) and will turn off
        chips that are not being used.

        Args:
            channels_to_read (List[int]): List of channel numbers to read
        """
        self._validate_channels_or_raise(channels_to_read, min_valid=0)
        for chan in range(self.board.channels):
            self._set_channel_active(chan, chan in channels_to_read)

    def number_events_to_read(self, amount: int) -> None:
        """Tell the board the maximum number of events to readout.

        This only works if the readout is set with singleEv is set to True.

        Args:
            amount (int): Maximum number of events to read.
        """
        max_winds = self.board.params["max_numwinds"]
        read_windows = self.board.registers["digital_registers"]["readoutwindows"][
            "value"
        ][0]
        channels = len(self.get_readout_channels())
        if not isinstance(amount, int):
            raise TypeError(f"Amount must be an integer, got {type(amount)}")

        numwinds_to_read = amount * read_windows * channels
        if not 0 <= numwinds_to_read <= max_winds:
            raise ValueError(
                f"Calculated number of windows {numwinds_to_read} is larger "
                f"than the maximum of {max_winds}."
            )

        LOGGER.debug(f"Setting numwinds to {numwinds_to_read}")
        self._write_control_register("numwinds", numwinds_to_read)

    def get_readout_channels(self) -> list:
        """Get the current channels specified as enabled in the read out.

        Returns:
            The list of channels read out.
        """
        return [
            chan
            for chan in range(self._board.channels)
            if self._board.registers["analog_registers"][f"ch_{chan}_en"]["value"][0]
        ]

    def _set_channel_active(self, channel: int, active: bool):
        """Wrapper for an analog register write that connects/disconnects
        the channel to everything, essentially turning it on of off

        Args:
            board (Board):
            channel (int): channel to act on
            active (bool): whether the channel should be turned on.

        Raises:
            TypeError if "channel" is not an int, or "active" not a bool.
            ValueError if the channel is out of bounds.
        """
        self._validate_channels_or_raise([channel])
        if not isinstance(active, bool):
            raise TypeError('"active" must be a bool')
        self._write_analog_registers(
            {
                f"bufbias_bias_{channel}": 0x3E8 if active else 0,
                f"ch_{channel}_en": active,
                f"ch_{channel}_isel_en": active,
            }
        )

    def _write_analog_registers(self, registers: dict, chip=None):
        """Write several analog registers.

        Args:
            registers (dict): a dict of `reg_name: reg_value`.
            chip (int): the chip number (0, 1). If not `None`, a suffix
                '_left' or '_right' is appended to each register name.
        """
        ar = AnalogRegisters(self.board)
        chip_suffix = {
            0: "_left",
            1: "_right",
        }
        for register, value in registers.items():
            ar.write(register.format(chip_suffix.get(chip, "")), value)

    def _compute_write_after_trig(self, write_after_trig):
        """On HDSoC write after trigger is equal to windows"""
        return write_after_trig



from logging import getLogger

from naludaq.communication import DigitalRegisters, AnalogRegisters, ControlRegisters
from naludaq.controllers.board.default import BoardController
from naludaq.helpers.semiton import SemitonABC

from typing import Iterable
import time


LOGGER = getLogger("naludaq.hdsocv2_trigger_controller")

BROADCAST = 64


class HDSoCv2BoardController(BoardController, SemitonABC):
    """Board controller for HDSoCv2.

    Only one instance of this class per board object is allowed.
    """

    @property
    def active_channels(self) -> list[int]:
        return self.board.params.get("active_channels", list())

    @active_channels.setter
    def active_channels(self, chans: list[int]):
        if not isinstance(chans, list):
            raise TypeError(f"active_channels must be a list, got {type(chans)}")
        if not all([isinstance(x, int) for x in chans]):
            raise TypeError("All channel numbers must be int")
        if not all(0 <= x < self.board.channels for x in chans):
            raise ValueError(
                f"all channels must be between 0 - {self.board.channels - 1}"
            )
        self.board.params["active_channels"] = chans

    def stop_readout(self):
        """Toggles the "stopacq" signal on the readout module.

        It's equivalent of asking it nicely to stop reading.
        """
        self.is_reading_out = False
        super().stop_readout()
        self.clear_buffer()

    def activate_channels(self, channels: list[int], dac_en_val: int = 1000):
        """Activate channels by powering the DACs and biases.

        Channels not specified will de diabled.

        Args:
            channels: list of channels to activate
            dac_en_val: value to set biases
        """
        dr = DigitalRegisters(self.board)
        ar = AnalogRegisters(self.board)
        prev_chan = self.active_channels
        for channel in range(self.board.channels):
            dr.write("selectchannel", channel)
            if (channel in channels) and (channel in prev_chan):
                continue
            elif channel in channels:
                active = True
            elif channel in prev_chan:
                active = False
            else:
                continue
            side = "left" if channel < 32 else "right"
            ar.write(f"channel_dac_bias_{side}", dac_en_val if active else 0)
            ar.write(f"channel_dac_bias_bias_{side}", dac_en_val if active else 0)
            ar.write(f"ch_fwd_{channel}", 6 if active else 2)
            ar.write(f"ch_rear_{channel}", 0x35 if active else 0x15)

        self.active_channels = channels
        dr.write("selectchannel", BROADCAST)

    def clear_buffer(self):
        """Clears the UART buffer on both CPU and FPGA side."""
        self._clear_fpga_buffer()
        super().clear_buffer()

    def _clear_fpga_buffer(self):
        """Resets the FPGA FIFO"""
        ControlRegisters(self.board).write("wave_fifo_rst", True)
        ControlRegisters(self.board).write("wave_fifo_rst", False)
        self.digital_reset()

    def read_scalers(self, channels: "Iterable[int] | None" = None) -> dict[int, int]:
        """Reads and returns all the digital scalar registers.

        The scalar registers are in two locations, one is the scal{ch} where the lower bits are
        stored, once that register is read, the scalhigh register is populated with the high bits.

        Args:
            channels (list[int]): list of channels to read scalars for.
                If not provided, scalars for all channels are read.

        Returns:
            list[int]: list of register read results for the channels selected.
                The list will not include values for disabled channels.

        Raises:
            ValueError: if a channel number in the list is invalid.
            TypeError: if channels is not a list or channel number is not an integer.
        """
        total_channels = self._get_total_channels()

        if channels is None:
            channels = range(total_channels)
        self._validate_channels_or_raise(channels)
        if self.is_reading_out:
            self.stop_readout()

        result = self._read_scalers(channels)
        return result

    def _read_scalers(self, channels: Iterable[int]) -> dict[int, int]:
        """Reads the scalars for the given channels, returns 0 for channels not selected."""
        result = {}
        for chan in channels:
            scalar = self.read_scalar(chan)
            result[chan] = scalar
        return result

    def read_scalar(self, channel: int, pause=0.001) -> int:
        # Select channel
        self._write_digital_register("selectchannel", channel)
        time.sleep(pause)
        # read bottom bits
        try:
            scal = self._read_digital_register("scal")
        except (KeyError, AttributeError) as e:
            LOGGER.warning("Can't read digital register scal due to: %s", e)
            scal = 0
        time.sleep(pause)
        # read top bits
        try:
            scalhigh = self._read_digital_register("scalhigh")
        except (KeyError, AttributeError) as e:
            LOGGER.warning("Can't read digital register scalhigh due to: %s", e)
            scalhigh = 0
        time.sleep(pause)
        # shift top bits up and add
        shift_amt = DigitalRegisters(self.board).registers["scal"]["bitwidth"]
        scal += scalhigh << shift_amt

        return scal



import functools
import itertools
import typing

from naludaq.helpers.exceptions import IterationError
from naludaq.helpers.validations import (
    validate_callable_or_raise,
    validate_non_negative_int_or_raise,
    validate_positive_int_or_raise,
)

_none = object()


class FancyIterator:
    def __init__(self, inner: typing.Iterable):
        """Initialize a new FancyIterator object wrapping the given iterable.

        This is a Rust-style iterator which follows a builder pattern for
        chaining operations together.

        Example:

        .. code-block:: python

            result = FancyIterator(range(10))
                .filter(lambda x: x % 2 == 0)
                .map(lambda x: x * 2)
                .collect()
            assert result == [0, 4, 8, 12, 16]

        Args:
            inner: An iterable object to be wrapped by the FancyIterator.
        """
        if not isinstance(inner, typing.Iterable):
            raise TypeError("wrapped value must be an iterable")
        self._inner = inner

    def __iter__(self):
        """Return the wrapped iterable."""
        return self._inner

    def copy(self) -> "FancyIterator":
        """Return a new iterator with the same wrapped iterable.

        This may have unexpected results if the wrapped iterable mutates data
        or is not re-usable.
        """
        return FancyIterator(self._inner)

    def inner(self) -> typing.Iterable:
        """Return the wrapped iterable."""
        return self._inner

    # =================================================================================
    # Exhausting Functions
    # =================================================================================
    def collect(self) -> list:
        """Return a list of all elements in the wrapped iterable.

        Important: this will attempt to exhaust the wrapped iterable; if it is an infinite iterable,
        this will never return and will blow up your memory.

        If a different datatype is desired, use this class as an iterator instead:
        ```
        deque(FancyIterator(inner).take(10))
        ```
        """
        return list(self)

    def count(self) -> int:
        """Exhaust the iterator and return the number of elements in the wrapped iterable.

        Important: this will attempt to exhaust the wrapped iterable; if it is an infinite iterable,
        this will never return and will blow up your memory.
        """
        try:
            return self.enumerate().last()[0] + 1
        except IterationError:
            return 0

    def last(self) -> typing.Any:
        """Exhaust the iterator and return the last element

        Raises:
            IterationError: if the wrapped iterable is empty.
        """
        no_last = object()
        last = no_last
        for x in self._inner:
            last = x
        if last is no_last:
            raise IterationError("last() called on empty iterator")
        return last

    def reduce(self, f: typing.Callable, initial: typing.Any = _none) -> typing.Any:
        """Exhaust the wrapped iterable and reduce it to a single value.

        The provided function is called pair-wise on the elements of the wrapped iterable in order,
        optionally starting with the provided initial value.

        Args:
            f (obj, obj -> obj): a callable that takes two elements of the wrapped iterable and returns a new element.
            initial (obj): the initial value to use for the reduction. If not provided, the first element of the
                wrapped iterable is used as the initial value.
        """
        validate_callable_or_raise(f)
        if initial is _none:
            return functools.reduce(f, self._inner)
        else:
            return functools.reduce(f, self._inner, initial)

    # =================================================================================
    # Iteration Stages
    # =================================================================================
    def accumulated(self) -> "FancyIterator":
        """Return a new iterator that yields accumulated elements of the wrapped iterable.

        All previous elements are yielded in a list.
        """
        return FancyIterator(_accumulate_inner(self._inner))

    def enumerate(self) -> "FancyIterator":
        """Return a new iterator that yields (index, element) tuples of the wrapped iterable."""
        return FancyIterator(enumerate(self._inner))

    def filter(
        self, f: typing.Callable, exclusion_limit: int = None
    ) -> "FancyIterator":
        """Return a new iterator that yields only elements that satisfy the given predicate.

        Args:
            f (obj -> bool): a callable that takes an element of the wrapped iterable and returns a boolean.
            exclusion_limit (int): the number of consecutive filtered elements to allow before raising an error.
                If None, no limit is applied.
        """
        validate_callable_or_raise(f)
        if exclusion_limit is not None:
            validate_positive_int_or_raise(exclusion_limit)
        return FancyIterator(_filter_inner(self._inner, f, exclusion_limit))

    def for_each(self, f: typing.Callable) -> "FancyIterator":
        """Apply a function to each element of the wrapped iterable.

        Args:
            f (obj -> None): a callable that takes an element of the wrapped iterable.
        """
        validate_callable_or_raise(f)
        return FancyIterator(_for_each_inner(self._inner, f))

    def map(self, f: typing.Callable) -> "FancyIterator":
        """Return a new iterator that maps elements of the wrapped iterable to new values.

        Args:
            f (obj -> obj): a callable that takes an element of the wrapped iterable and returns a new element.
        """
        validate_callable_or_raise(f)
        return FancyIterator(map(f, self._inner))

    def skip(self, count: int) -> "FancyIterator":
        """Return a new iterator that skips the first `count` elements of the wrapped iterable."""
        validate_non_negative_int_or_raise(count)
        return FancyIterator(itertools.islice(self._inner, count, None))

    def skip_while(self, f: typing.Callable) -> "FancyIterator":
        """Return a new iterator that skips elements of the wrapped iterable as long as the predicate is satisfied.

        Args:
            f (obj -> bool): a callable that takes an element of the wrapped iterable and returns a boolean.
        """
        validate_callable_or_raise(f)
        return FancyIterator(itertools.dropwhile(f, self._inner))

    def take(self, count: int) -> "FancyIterator":
        """Return a new iterator that yields only the first `count` elements of the wrapped iterable."""
        validate_non_negative_int_or_raise(count)
        return FancyIterator(itertools.islice(self._inner, count))

    def take_while(self, f: typing.Callable) -> "FancyIterator":
        """Return a new iterator that yields elements of the wrapped iterable as long as the predicate is satisfied."""
        validate_callable_or_raise(f)
        return FancyIterator(itertools.takewhile(f, self._inner))

    def unaccumulated(self) -> "FancyIterator":
        """Return a new iterator that yields only the last element of each accumulated list.

        Raises:
            IterationError: on iteration if the wrapped iterable is not an accumulated iterator.
        """
        return self.map(_unaccumulate_inner)

    def unenumerate(self) -> "FancyIterator":
        """Return a new iterator that yields only the element portion of each (index, element) tuple.

        Raises:
            IterationError: on iteration if the wrapped iterable is not an enumerated iterator.
        """
        return self.map(_unenumerate_inner)

    def zip(self, *others) -> "FancyIterator":
        """Return a new iterator that yields tuples of elements from the wrapped iterable and other iterables."""
        if any(not isinstance(x, typing.Iterable) for x in others):
            raise TypeError("all arguments must be iterable")
        return FancyIterator(zip(self._inner, *others))


def _accumulate_inner(iter):
    x = []
    for e in iter:
        x.append(e)
        yield x


def _filter_inner(iter, f, exclusion_limit):
    filter_count = 0
    for e in iter:
        if f(e):
            filter_count = 0
            yield e
        else:
            filter_count += 1
            if exclusion_limit is not None and filter_count > exclusion_limit:
                raise IterationError(
                    f"filter limit of {exclusion_limit} consecutive elements reached"
                )


def _for_each_inner(iter, fn):
    for e in iter:
        fn(e)
        yield e


def _unaccumulate_inner(x):
    if not isinstance(x, list) or len(x) == 0:
        raise IterationError("unaccumulated() called on non-accumulated iterator")
    return x[-1]


def _unenumerate_inner(x):
    if not isinstance(x, tuple) or len(x) != 2 or not isinstance(x[0], int):
        raise IterationError("unenumerate() called on non-enumerated iterator")
    return x[1]



"""Initializer for the HDSoCv1 board

This initializer will run the init sequence for the HDSoCv1 board.
It contains the HDSoCv1 specific startup sequence.
"""
import logging
import time

from naludaq.board.initializers import Initializers
from naludaq.helpers.decorators import log_function

logger = logging.getLogger("naludaq.init_hdsocv1")

VALID_BOARDS = [
    "hdsocv1",
    "hdsocv1_evalr1",
    "hdsocv1_evalr2",
]


class InitHDSoCv1(Initializers):
    def __init__(self, board):
        """Initializer for HDSoCv1.

        Args:
            board (Board): the board to initialize.
        """
        super().__init__(board, VALID_BOARDS)
        self.power_sequence = [
            "2v5_en",
            "1v2_en",
            "3v3_i2c_en",
            "clk2v5_en",
            "clk1v8_en",
            "clk_i2c_sel",
        ]

    def run(self) -> bool:
        """Runs the initialization sequence.

        Returns:
            True, always.
        """
        # Initialize FPGA registers
        self._fpga_init()

        # Power on FMC board
        self._power_toggle(False)
        self._power_toggle(True)
        time.sleep(0.25)

        self._program_clock()

        self._set_serial_mode(self.serial_mode_en)
        time.sleep(0.1)
        self._set_loopback_enabled(False)

        # Chip-side register startup
        self._analog_startup()
        self._digital_startup()

        # Set DACs
        self._init_dacs()

        return True

    @property
    def serial_mode_en(self):
        return self.board.params.get("serial_mode", False)

    @log_function(logger)
    def _fpga_init(self):
        """Write all FPGA registers to their default values."""
        self.control_registers.write_all()
        self.i2c_registers.write("i2c_en", True)

    @log_function(logger)
    def _power_toggle(self, state):
        """Toggles the power rails defined in `power_sequence` attribute."""
        for register in self.power_sequence:
            self.control_write(register, state)

    @log_function(logger)
    def _set_serial_mode(self, enabled: bool):
        """Turn on or off serial mode."""
        self.board.control.set_serial_mode(enabled)

    @log_function(logger)
    def _set_loopback_enabled(self, enabled: bool):
        """Enable or disable loopback mode."""
        self.board.control.set_loopback_enabled(enabled)

    @log_function(logger)
    def _program_clock(self):
        """Program the clock chip, uses the clockfile in the paramters if it exits."""
        self.board.clock.program(filename=self.board.params["clock_file"])
        self.control_write("clk_oeb", False)
        self.control_write("sysrst", True)
        self.control_write("sysrst", False)
        self.control_write("regclr", True)
        self.control_write("regclr", False)

    @log_function(logger)
    def _reset_sys(self):
        """Reset the ASIC and clear the analog and digital registers"""
        self.control_write("sysrst", True)
        self.control_write("regclr", True)

        self.control_write("sysrst", False)
        self.control_write("regclr", False)

    @log_function(logger)
    def _digital_startup(self):
        """Startup the digital side of the chip by programming all registers."""
        self.digital_registers.write_all()

    @log_function(logger)
    def _analog_startup(self):
        """Start the analog side of the chip by programming all registers."""
        self.analog_registers.write_all()
        self._dll_startup()

    @log_function(logger)
    def _dll_startup(self):
        """Starting the delay line on the analog side.

        Sets and unsets vanbuff to get the dll going and
        changes the vadjp values to ensure proper SST duty cycle once locked.
        """

        self.analog_write("qbias_left", 0)
        self.analog_write("qbias_right", 0)
        self.analog_write("vanbuf_left", 2816)
        self.analog_write("vanbuf_right", 2816)
        time.sleep(1)
        self.analog_write("qbias_left", 2048)
        self.analog_write("qbias_right", 2048)
        self.analog_write("vanbuf_left", 0)
        self.analog_write("vanbuf_right", 0)

        self._set_vadjp()

    def _set_vadjp(self):
        try:
            vadjp_right = self.board.registers["analog_registers"]["vadjp_right"][
                "value"
            ]
            vadjp_left = self.board.registers["analog_registers"]["vadjp_left"]["value"]
            if isinstance(vadjp_right, list):
                vadjp_right = vadjp_right[0]
            if isinstance(vadjp_left, list):
                vadjp_left = vadjp_left[0]
            logger.debug(f"vadjp_right: {vadjp_right}, vadjp_left: {vadjp_left}")
        except KeyError:
            raise KeyError(
                "vadjp_right or vadjp_left not found in the analog_registers"
            )
        self.analog_write("vadjp_right", vadjp_right)
        self.analog_write("vadjp_left", vadjp_left)

    @log_function(logger)
    def _init_dacs(self):
        """Program the left and right side DACs"""
        dacvalues = self._get_dac_values()

        for channel, value in dacvalues.items():
            self.board.ext_bias.set_single_dac(channel, value)

    def _get_dac_values(self) -> dict:
        """Return a dict with all the {channels: dac_values}"""
        return self.board.params.get("ext_dac", {}).get("channels", {})



from .asoc import ASoCHeaderParser
from .asocv2 import ASoCv2HeaderParser
from .base import NoOpHeaderParser
from .siread import SiREADHeaderParser
from .trbhm import TrbhmHeaderParser
from .upac32 import UPAC32HeaderParser

_HEADER_PARSERS = {
    "aardvarcv2": ASoCHeaderParser(),
    "aardvarcv3": ASoCHeaderParser(),
    "aodsv1": ASoCv2HeaderParser(),
    "aodsv2_eval": ASoCv2HeaderParser(),
    "asoc": ASoCHeaderParser(),
    "asocv2": ASoCv2HeaderParser(),
    "asocv3": ASoCv2HeaderParser(),
    "hdsocv1": ASoCHeaderParser(),
    "hdsocv1_evalr1": ASoCHeaderParser(),
    "hdsocv1_evalr2": ASoCHeaderParser(),
    "hiper": ASoCHeaderParser(),
    "siread": SiREADHeaderParser(),
    "trbhm": TrbhmHeaderParser(),
    "dsa-c10-8": TrbhmHeaderParser(),
    "udc16": NoOpHeaderParser(),
    "upac32": UPAC32HeaderParser(),
    "upaci": UPAC32HeaderParser(),
    "zdigitizer": UPAC32HeaderParser(),
    "default": NoOpHeaderParser(),
}


def get_header_parser(model: str):
    """Gets a header parser function appropriate for the given board.

    Args:
        model (str): the board model

    Returns:
        The header parser function.
    """
    return _HEADER_PARSERS.get(model, NoOpHeaderParser()).parse_header



import numpy as np

from naludaq.communication import AnalogRegisters
from naludaq.controllers import get_dac_controller
from naludaq.helpers import validations
from naludaq.helpers.exceptions import OperationCanceledError
from naludaq.helpers.operations import CancelableOperation, ProgressReporter
from naludaq.tools import get_data_collector

from .bayesian_optimizer import BayesianOptimizer


class GainStageTuner(ProgressReporter, CancelableOperation):
    def __init__(
        self,
        board,
        isel_bounds: tuple[int, int],
        dac_bounds: tuple[int, int],
        cost_function=None,
    ):
        """Tunes the DAC and ISEL values of the board to center the ADC values around midrange for each channel.

        This tool is intended to be used for finding the ideal DAC and ISEL values in order to center the ADC values
        around midrange (2048 ADC counts on 12-bit ADCs). It uses Bayesian optimization to find the DAC and ISEL values
        that give the best cost values independently for each chip. The cost value is calculated by the cost function,
        which by default calculates the difference from the average to midrange for each channel.
        The DAC and ISEL values are set to the values that gave the best cost values upon completion.

        Care should be taken when selecting the ISEL bounds to search within, as ISEL values outside of a certain range
        can damage certain chips.

        Args:
            board (Board): Board to tune. Needs an open connection
            isel_bounds (tuple[int, int]): Minimum and maximum ISEL values.
            dac_bounds (tuple[int, int]): Minimum and maximum DAC values.
            cost_function (function): Function to calculate the cost for a given set of DAC and ISEL values.
                Defaults to None, which scores the costs values based on the difference from the average to midrange (2048).
                Any function with the signature `fn(acq: list[dict], chip: int) -> float` can be used.
        """
        super().__init__()
        self._board = board
        self._channels_per_chip = self._board.channels // self._board.available_chips
        self._midrange = 2 ** (self._board.params.get("resolution", 12) - 1)

        self._bounds = {}
        self.isel_bounds = isel_bounds
        self.dac_bounds = dac_bounds
        self.cost_function = cost_function or self._calculate_mean_diff

        self.events_per_probe = 10
        self.attempts_per_event = 10
        self.readout_params = {"trig": "ext", "lb": "forced"}
        self.read_window = {
            "windows": 10,
            "lookback": 20,
            "write_after_trig": 254,
        }

    @property
    def board(self):
        """Get the board being tuned."""
        return self._board

    @property
    def cost_function(self):
        """Get/set the function to calculate the cost value for a given set of DAC and ISEL values.

        The function must have the signature `fn(acq: list[dict], chip: int) -> float`.
        """
        return self._cost_function

    @cost_function.setter
    def cost_function(self, fn):
        validations.validate_callable_or_raise(fn)
        self._cost_function = fn

    @property
    def isel_bounds(self) -> tuple[int, int]:
        """Get/set the minimum and maximum ISEL values.

        Care should be taken when selecting the ISEL bounds to search within, as ISEL values outside of a certain range
        can damage certain chips.
        """
        return self._bounds["y"]

    @isel_bounds.setter
    def isel_bounds(self, value: tuple[int, int]):
        validations.validate_positive_int_range_or_raise(value)
        self._bounds["y"] = tuple(value)

    @property
    def dac_bounds(self) -> tuple[int, int]:
        """Get/set the minimum and maximum DAC values."""
        return self._bounds["x"]

    @dac_bounds.setter
    def dac_bounds(self, value: tuple[int, int]):
        validations.validate_positive_int_range_or_raise(value)
        value = tuple(value)
        low, high = value
        max_counts = self.board.params["ext_dac"]["max_counts"]
        min_counts = self.board.params["ext_dac"].get("min_counts", 0)
        if low < min_counts:
            raise ValueError(f"Lower bound must be greater than {min_counts}")
        if high > max_counts:
            raise ValueError(f"Upper bound must be less than {max_counts}")
        self._bounds["x"] = value

    @property
    def events_per_probe(self) -> int:
        """Get/set the number of events to collect per probe.

        In this context "probe" means a single set of DAC and ISEL values.
        """
        return self._events_per_probe

    @events_per_probe.setter
    def events_per_probe(self, value):
        validations.validate_positive_int_or_raise(value)
        self._events_per_probe = value

    @property
    def attempts_per_event(self) -> int:
        """Get/set the number of attempts to collect an event."""
        return self._attempts_per_event

    @attempts_per_event.setter
    def attempts_per_event(self, value):
        validations.validate_positive_int_or_raise(value)
        self._attempts_per_event = value

    @property
    def readout_params(self) -> dict:
        """Get/set the readout parameters."""
        return self._readout_params

    @readout_params.setter
    def readout_params(self, value: dict):
        validations.validate_readout_settings(value)
        self._readout_params = value

    @property
    def read_window(self) -> dict:
        """Get/set the read window parameters."""
        return self._read_window

    @read_window.setter
    def read_window(self, value: dict):
        validations.validate_read_window_dict_or_raise(value)
        self._read_window = value

    def run(self, n_iter=50) -> list[tuple]:
        """Runs the optimization.

        Upon completion, the DAC and ISEL values of the board will be set to the values
        that gave the best cost values.

        Args:
            n_iter (int): Number of iterations to run. Defaults to 50.

        Returns:
            list[tuple]: List of tuples with the DAC and ISEL values for each chip.
                The first element of the tuple is the DAC value, the second is the ISEL value.

        Raises:
            OperationCanceledError: If the operation has been canceled.
            TimeoutError: If data capture timed out.
        """
        optimizers = self._create_optimizers()
        for i in range(n_iter):
            self._raise_if_cancelled()
            self.update_progress(
                (i / n_iter) * 100, f"Running optimization ({i+1}/{n_iter})"
            )
            self._run_iteration_or_raise(optimizers)

        results = [opt.maximum["params"] for opt in optimizers]
        results = self._round_params(results)
        self._apply_suggestions(results)
        return [(res["x"], res["y"]) for res in results]

    def _run_iteration_or_raise(self, optimizers: list[BayesianOptimizer]):
        """Runs a single iteration of the optimization.

        Args:
            optimizers (list[BayesianOptimizer]): List of optimizers to use.
        """
        suggestions = [optimizer.suggest() for optimizer in optimizers]
        suggestions = self._round_params(suggestions)
        self._apply_suggestions(suggestions)
        try:
            data = self._get_data()
        except (OperationCanceledError, TimeoutError):
            raise
        for chip, optimizer in enumerate(optimizers):
            cost = self._cost_function(data, chip)
            optimizer.register(suggestions[chip], cost)

    def _apply_suggestions(self, points: list[dict]):
        """Writes the suggested DAC and ISEL values to the board.

        Args:
            points (list[dict]): List of dictionaries with "x" and "y" keys, where x is the DAC value
                and y is the ISEL value.
        """
        for chip, point in enumerate(points):
            self._set_dac(chip, point["x"])
            self._set_isel(chip, point["y"])

    def _create_optimizers(self):
        """Creates the BayesianOptimizer instances for each chip."""
        return [
            BayesianOptimizer(cost_function=None, bounds=self._bounds)
            for _ in range(self.board.available_chips)
        ]

    def _get_data(self) -> list[dict]:
        """Collects data from the board.

        Uses the readout parameters set in the constructor.

        Returns:
            list[dict]: list of events

        Raises:
            OperationCanceledError: If the operation has been canceled.
            TimeoutError: If the operation timed out.
        """
        collector = get_data_collector(self._board)
        collector.set_window(**self._read_window)
        collector.readout_settings = self._readout_params
        data = (
            collector.iter(
                count=self._events_per_probe,
                attempts=self._attempts_per_event,
            )
            .for_each(lambda _: self._raise_if_cancelled)
            .collect()
        )
        return data

    def _calculate_mean_diff(self, acq: list[dict], chip: int) -> float:
        """Calculates the difference from the acq's average to midrange (2048).
        Calculates the cost value for the channels of the given chip

        Args:
            acq (list[dict]): List of events.
            chip (int): Chip number to calculate the cost value for.

        Returns:
            float: The cost value for the given chip.
        """
        evt_averages = []
        for evt in acq:
            evt_averages.append(
                [np.mean(chandata) for chandata in evt["data"] if len(chandata) != 0]
            )

        chan_averages = np.mean(evt_averages, axis=0)

        # compute cost value
        cpc = self._channels_per_chip
        chan_averages = chan_averages[cpc * chip : cpc * chip + cpc]
        cost_value = 0 - np.sum([np.abs(avg - self._midrange) for avg in chan_averages])

        return cost_value

    def _round_params(self, params: list[dict]) -> dict:
        """Rounds the DAC and ISEL values in the given parameters dictionaries.

        Float values are rounded to the nearest integer.

        Args:
            params (list[dict]): List of dictionaries with "x" and "y" keys, where x is the DAC value
                and y is the ISEL value.

        Returns:
            dict: Dictionary with the rounded DAC and ISEL values.
        """
        return [{k: round(v) for k, v in d.items()} for d in params]

    def _set_isel(self, chip: int, isel: int):
        """Sets the ISEL value for a chip.

        Args:
            chip (int): Chip number to set the ISEL value for.
            isel (int): ISEL value to set.
        """
        AnalogRegisters(self._board, chip).write("isel", isel)

    def _set_dac(self, chip: int, counts: int):
        """Sets the DAC values for a chip.

        Args:
            chip (int): Chip number to set the DAC values for.
            counts (int): DAC value to set.
        """
        start_chan = chip * self._channels_per_chip
        end_chan = start_chan + self._channels_per_chip
        get_dac_controller(self._board).set_dacs(
            counts, list(range(start_chan, end_chan))
        )



"""
"""
import time
from logging import getLogger

from naludaq.backend.managers.connection import ConnectionManager
from naludaq.backend.managers.io import BoardIoManager
from naludaq.backend.models.device import DeviceType
from naludaq.communication import ControlRegisters
from naludaq.controllers.controller import Controller

LOGGER = getLogger("naludaq.board_controller_udc")
NON_UDP = [DeviceType.SERIAL, DeviceType.D2XX, DeviceType.D3XX]


class UDCBoardController(Controller):
    """Board cotnroller for the specific UDC functionality.

    The UDC board reads out all data, every time.
    It has so far very limited readout modes.

    """

    def __init__(self, board):
        super().__init__(board)
        self._write_pause = 0.01

    def start_readout(
        self,
        trigger_type="self",
    ):
        """Sends a readout signal to the board.

        Obstensibly reads out the chip?

        Args:
            trigger_type (str):
                "There are 4 trigger modes:

                '00'	: 	Software Trigger Mode, the power on default.
                '01'	: 	Auto Trigger Mode. 1-2Hz auto trigger data rate.
                '10'	:	External Trigger Mode.
                '11'	: 	Self Trigger Mode.

        Command
        -------

        command is: BMMXXVVV
        MM = singleEvent(7) + readoutEn(6) + pedSel(5 dt 4) + modeSel(3 dt 0)
        - modeSel=x"0" analog write
        - modeSel=x"1" digital write
        - modeSel=x"2" digital read
        - modeSel=x"3" waveform read

        VVVVV = addr_sel(19 dt 12) + data_sel(11 dt 0)
        - data_sel(5 downto 4) = trig type
        - data_sel(3 downto 2) = lb type
        - data_sel(1 downto 0) = acq type

        """
        self.is_reading_out = True
        self._enable_rx_tx(False)
        self.sysrst(regclr=False)
        self.digital_reset()

        # # wave fifo reset
        self._clear_fpga_buffer()
        self._enable_rx_tx(True)

    def toggle_trigger(self):
        """Toggles the software trigger bit, should result in a software trigger event."""
        LOGGER.debug("Toggle software trigger")
        self._enable_rx_tx(False)
        self.sysrst(regclr=False)
        self.digital_reset()

        # # wave fifo reset
        self._clear_fpga_buffer()
        self._enable_rx_tx(True)
        self.arm()
        cmd = "C0000064"  # Send software trigger
        self._send_command(cmd)
        self.toggle_reread()

    def arm(self):
        """Arms the system by toggling the "udc_arm" register"""
        self._write_control_register("udc_arm", True)
        self._write_control_register("udc_arm", False)

    def toggle_reread(self):
        """"""
        self._write_control_register("reread_data", True)
        self._write_control_register("reread_data", False)

    def reset_board(self):
        """Try and reset the board.

        In case the FPGA get stuck this can help reset the state.
        """
        time.sleep(0.02)
        self.digital_reset()
        time.sleep(0.02)
        self.sysrst()

    def digital_reset(self):
        """Toggles the "reset" port on the readout module.

        Forcibly returns the chip to default state.
        """
        self._write_control_register("digrst", True)
        self._write_control_register("digrst", False)

    def clear_buffer(self):
        """Clears the UART buffer on both CPU and FPGA side."""
        try:
            self._clear_fpga_buffer()
        except Exception:
            pass

        try:
            self._clear_input_buffer()
        except Exception:
            pass

    def _clear_fpga_buffer(self):
        self._write_control_register("wave_fifo_reset", True)
        self._write_control_register("wave_fifo_reset", False)

    def _clear_input_buffer(self):
        if self.board.using_new_backend:
            device = ConnectionManager(self.board).device
            if device.type in NON_UDP:
                device.clear_buffers()
        else:
            self.board.connection.reset_input_buffer()

    def sysrst(self, regclr: bool = True):
        """Toggles the sysrst pin, which resets the digital portion of the chip

        Args:
            regclr (bool): whether to toggle the 'regclr' register.
        """
        self._write_control_register("sysrst", True)
        regclr and self._write_control_register("regclr", True)
        self._write_control_register("sysrst", False)
        regclr and self._write_control_register("regclr", False)

    def _enable_rx_tx(self, en: bool):
        """Turns on or off serial interface Rx and Tx.

        Args:
            en (bool): whether to enable Rx and Tx.
        """
        self._write_control_register("rx_enable", en)
        self._write_control_register("tx_enable", en)

    #############################################################################
    # Random Stuff, may or maynot work... Below functions need to be validated.
    #############################################################################

    def read_scalers(self, channels: list[int] = None):
        """Not implemented."""
        # self.stop_readout()
        scalers = []

        return scalers

    def enable_testmode(self, enabled: bool):
        """Enables the test-pattern output.

        Set the fpga in a mode to output a known test-pattern.
        """
        # ControlRegisters(self.board).write("fake", enabled)

    def get_available_chips(self) -> list[int]:
        """Get a list of available chip numbers."""
        return list(range(self.board.params.get("num_chips", 1)))

    def _write_control_register(self, register, value):
        """wrapper for the Control register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        """
        try:
            ControlRegisters(self.board).write(register, value)
            time.sleep(self._write_pause)
        except (ValueError, TypeError) as error_msg:
            LOGGER.error("Couldn't update control register due to: %s", error_msg)

    def _send_command(self, command):
        """Send the given hex command to the board.

        Args:
            command (str): hex command
        """
        if self.board.using_new_backend:
            BoardIoManager(self.board).write(command)
        else:
            self.board.connection.send(command)



"""Thresholdscanner

"""


def get_threshold_scanner(board, *args, **kwargs):
    """Gets a `ThresholdScan` object appropriate for a given board.

    Args:
        board (Board): the board object.

    Returns:
        The ThresholdScan object which can be used to run threshold
        scans on the board.
    """
    if board.model in ["hdsocv1_evalr1", "hdsocv1_evalr2"]:
        from naludaq.tools.threshold_scan.hdsoc_thresholdscan import ThresholdScanHdsoc

        return ThresholdScanHdsoc(board, *args, **kwargs)
    elif board.model in [
        "hdsocv2_eval",
        "hdsocv2_evalr2",
    ]:
        from naludaq.tools.threshold_scan.hdsocv2_thresholdscan import (
            ThresholdScanHdsocv2,
        )

        return ThresholdScanHdsocv2(board, *args, **kwargs)
    elif board.model in ["upac96"]:
        from naludaq.tools.threshold_scan.upac96_thresholdscan import (
            Upac96ThresholdScan,
        )

        return Upac96ThresholdScan(board, *args, **kwargs)
    elif not board.is_feature_enabled("threshold_scan"):
        raise NotImplementedError(f"Threshold scan not supported on {board.model}")
    else:
        from naludaq.tools.threshold_scan.threshold_scan import ThresholdScan

        return ThresholdScan(board, *args, **kwargs)



"""Helper functions for validating inputs."""
import pathlib
import typing


def validate_dir_or_raise(output_dir, name="Directory"):
    if not isinstance(output_dir, (str, pathlib.Path)):
        raise TypeError(f"{name} must be a string.")
    if not pathlib.Path(output_dir).exists():
        raise FileNotFoundError(f"{name} must exist.")
    if output_dir is None or not pathlib.Path(output_dir).is_dir():
        raise NotADirectoryError(f"{name} must be specified.")


def validate_chip_list_or_raise(chips: list[int], params: dict):
    num_chips = len(params.get("chips", ["single"]))
    if not isinstance(chips, list):
        raise TypeError("Chips must be a list.")
    if len(chips) != len(set(chips)):
        raise ValueError("One or more chip numbers is repeated.")
    for chip in chips:
        if not isinstance(chip, int):
            raise TypeError("Chips must be a list of integers.")
        if not 0 <= chip < num_chips:
            raise ValueError("One or more chip numbers is out of range.")


def validate_channel_sequence_or_raise(params: dict, channels: typing.Iterable[int]):
    max_channels = params["channels"]
    if not isinstance(channels, typing.Iterable):
        raise TypeError("Value must be iterable")
    if not all(isinstance(c, int) for c in channels):
        raise TypeError("All channels must be integers")
    if not all(0 <= c < max_channels for c in channels):
        raise ValueError(f"All channels must be in range 0-{max_channels - 1}")
    if len(channels) != len(set(channels)):
        raise ValueError("One or more channels is repeated.")


def validate_positive_int_or_raise(value: int):
    if not isinstance(value, int):
        raise TypeError(f"value must be an integer.")
    if value <= 0:
        raise ValueError(f"value must be positive.")


def validate_non_negative_int_or_raise(value: int):
    if not isinstance(value, int):
        raise TypeError(f"value must be an integer.")
    if value < 0:
        raise ValueError(f"value must be non-negative.")


def validate_callable_or_raise(fn: typing.Callable):
    # this is incomplete, but works for most callable objects
    if not callable(fn):
        raise TypeError("Value must be callable.")


def validate_readout_settings(readout_settings: dict):
    valid_readout_settings = {
        "trig": ["i", "s", "e"],
        "lb": ["t", "f", "r"],
        "acq": ["p", "r"],
        "ped": ["z", "c", "r"],
        "readoutEn": [True, False],
        "singleEv": [True, False],
    }

    # Validate values
    for key, value in readout_settings.items():
        expected = valid_readout_settings.get(key, None)
        test_val = value[0] if isinstance(value, str) else value
        if expected is None:
            raise KeyError(f"'{key}' is not a valid readout setting.")
        if test_val not in expected:
            raise KeyError(
                f"'{value}' is not a valid value for setting '{key}'. Valid values are {expected}"
            )


def validate_read_window_dict_or_raise(read_window: dict):
    """Validate the read window dictionary.

    Must contain a subset of the following keys: "windows", "lookback", "write_after_trig"
    Must contain only integer values.
    """
    valid_keys = ["windows", "lookback", "write_after_trig"]
    if not isinstance(read_window, dict):
        raise TypeError("Read window must be a dictionary.")
    if any(k not in valid_keys for k in read_window.keys()):
        raise KeyError(
            f"Read window must contain a subset of the following keys: {', '.join(valid_keys)}"
        )
    if any(not isinstance(v, int) for v in read_window.values()):
        raise TypeError("Read window values must be integers.")


def validate_positive_int_range_or_raise(endpts: "tuple | list"):
    """Validate a tuple or list specified as (low, high).

    The range must be positive.
    """
    if not isinstance(endpts, (tuple, list)):
        raise TypeError("range must be a tuple or list.")
    if len(endpts) != 2:
        raise ValueError("range must be specified as (low, high).")
    low, high = tuple(endpts)
    if not isinstance(low, int) or not isinstance(high, int):
        raise TypeError("range values must be integers.")
    if low > high:
        raise ValueError("range must be specified as (low, high).")
    if low < 0:
        raise ValueError("negative  are not allowed.")



"""

"""
import logging
import time

import numpy as np

from naludaq.communication.analog_registers import AnalogRegisters
from naludaq.controllers import get_readout_controller
from naludaq.helpers import type_name
from naludaq.tools.threshold_scan.threshold_scan import ThresholdScan

LOGGER = logging.getLogger("naludaq.threshold_scan")


class ThresholdScanHdsoc(ThresholdScan):
    """Tool for sweeping over trigger thresholds and counting number of triggers.
    Can be used to determine the ideal point to set the trigger.

    """

    def __init__(
        self,
        board,
        *args,
        low_ref_value: int = None,
        high_ref_value: int = None,
        **kwargs,
    ):
        """Threshold scan for hdsoc boards.

        Args:
            board (Board): board object
            start_value (int): the start value for the scan
            stop_value (int): the stop value for the scan
            step_size (int): the step size for the scan
            low_ref_value (int): low reference value
            high_ref_value (int): high reference value
        """
        super().__init__(board, *args, **kwargs)
        self.low_reference = low_ref_value
        self.high_reference = high_ref_value

    @property
    def _max_reference(self) -> int:
        """Get the maximum number of references allowed"""
        return self.board.trigger.params.get("max_ref", 15)

    @property
    def _min_reference(self) -> int:
        """Get the minimum reference value allowed"""
        return self.board.trigger.params.get("min_ref", 0)

    @property
    def low_reference(self) -> int:
        """Get/set the low trigger reference value used during the scan"""
        return self._low_ref_value

    @low_reference.setter
    def low_reference(self, value: int):
        if value is None:
            value = self.board.params.get("threshold_scan", {}).get("low_ref", 0)
        if not isinstance(value, int):
            raise TypeError(f"Value must be int, not {type_name(value)}")
        if not self._min_reference <= value <= self._max_reference:
            raise ValueError(
                f"Value {value} is out of bounds for range "
                f"{self._min_reference} - {self._max_reference}"
            )
        self._low_ref_value = value

    @property
    def references(self) -> tuple[int, int]:
        """Get/Set the references (low, high) used during the scan"""
        return (self.low_reference, self.high_reference)

    @references.setter
    def references(self, values: tuple[int, int]):
        self.low_reference, self.high_reference = values

    @property
    def high_reference(self) -> int:
        """Get/set the high trigger reference value used during the scan"""
        return self._high_ref_value

    @high_reference.setter
    def high_reference(self, value: int):
        if value is None:
            value = self.board.params.get("threshold_scan", {}).get("high_ref", 0)
        if not isinstance(value, int):
            raise TypeError(f"Value must be int, not {type_name(value)}")
        if not self._min_reference <= value <= self._max_reference:
            raise ValueError(
                f"Value {value} is out of bounds for range "
                f"{self._min_reference} - {self._max_reference}"
            )
        self._high_ref_value = value

    def run(self, pause: float = 0.1):
        """Reimplementation to perform extra validation"""
        self._validate_scan_settings_or_raise()
        return super().run(pause)

    def _get_scalar_values(self, pause: float = 0.1):
        """Reimplementation to perform some extra setup."""
        rc = get_readout_controller(self.board)
        rc.set_readout_channels(self.channels)
        tc = self.board.trigger
        tc.tsel = {
            "0_15": True,
            "16_31": True,
        }
        tc.references = {
            "0_15": self.references,
            "16_31": self.references,
        }
        return self._get_scalar_values_vertical(pause)

    def _backup_board_settings(self):
        """Back up the trigger values, readout channels, references, and tsel"""
        super()._backup_board_settings()
        rc = get_readout_controller(self.board)
        tc = self.board.trigger
        self._original_readout_channels = rc.get_readout_channels()
        self._original_references = tc.references
        self._original_tsel = tc.tsel

    def _restore_board_settings(self):
        """Restore the trigger values, readout channels, references, and tsel"""
        super()._restore_board_settings()
        rc = get_readout_controller(self.board)
        rc.set_readout_channels(self._original_readout_channels)
        tc = self.board.trigger
        tc.references = self._original_references
        tc.tsel = self._original_tsel

    def _set_tsel_enabled(self, left: bool, right: bool):
        """Set whether tsel is enabled on each side.

        During the threshold scan, tsel must be enabled for scalars
        to read correctly.
        """
        AnalogRegisters(self.board).write("tsel_left", left)
        AnalogRegisters(self.board).write("tsel_right", right)

    def _get_analog_register(self, register: str) -> int:
        """Get the value of an analog register"""
        return self.board.registers["analog_registers"][register]["value"][0]

    def _validate_scan_settings_or_raise(self):
        """Raise an error if scan settings don't make sense"""
        if self.low_reference >= self.high_reference:
            raise ValueError("Low reference cannot exceed the high reference")

    def _get_scalar_values_vertical(self, pause: float) -> np.ndarray:
        """Scan the range and return np.array with trigger amounts

        Args:
            pause (float): amount of seconds to pause in between samples.

        Returns:
            Triggers value per channel as `np.array`.
        """
        scan_values = self.scan_values
        output = np.zeros((self.board.channels, len(scan_values)))

        # NOTE: the default threshold scan iterates over scan values then channels,
        # but on HDSoC it is super important that this order is reversed!
        # Otherwise the threshold scan will be super messed up or not work at all.
        for ch_i, chan in enumerate(self.channels):
            for i, value in enumerate(scan_values):
                if self._cancel:
                    break
                progress = ch_i * len(scan_values) + i + 1
                progress_total = len(self.channels) * len(scan_values)
                LOGGER.debug("Scan progress: %s/%s", progress, progress_total)
                self._progress.append(
                    (
                        int(95 * progress / progress_total),
                        f"Scanning {progress}/{progress_total}",
                    )
                )
                default_tval = self.board.trigger.params.get("default_trigger_value", 0)
                tvalues = {ch: default_tval for ch in range(self.board.channels)}
                tvalues[chan] = int(value)
                self.board.trigger.values = tvalues  # int(value)
                time.sleep(pause)

                scalar = self._get_single_scalar_value(chan)
                output[chan][i] = scalar

        return output

    def _get_single_scalar_value(self, channel) -> list[int]:
        """Gets a reading from all scalers on the board.

        Returns:
            list[int]: scalar values ordered by channel.
        """
        return self._board_controller.read_scalar(channel)
